# -*- coding: utf-8 -*-
"""
Demofab example

To view predefined layers in klayout load:
lyp_demo.lyp
"""

import nazca as nd
import nazca.demofab.pdk as demo
from nazca.demopackager.packages import Package1


#==============================================================================
# Die 1
#==============================================================================
DIE = demo.DesignArea(name='Die1')
DIE.size(length=2500, height=2000)
DIE.cleave = 100
DIE.coating('AR', 'NO')

print('Die height:', DIE.height)
print('Die length:', DIE.length)


#==============================================================================
# Package 1
#==============================================================================
PACK = Package1(name='Pack1', length=DIE.length, height=DIE.height)
PACK.DCx = 250
PACK.DCy = 200
PACK.DCpitch = 300
PACK.DCcenter = True
PACK.RFx = 150
PACK.RFn = 4
PACK.RFcenter = True
PACK.fiberarea = 1000


#==============================================================================
# Utils
#==============================================================================
def addAllBBs(instance, BBcell, pinprefix):
    """Place BBs on all pins in an instance matching a string in pinprefix"""
    AllBBs = {name: BBcell.put('rc', p) for name, p in instance.pin.items() if pinprefix in name}

    return AllBBs


#==============================================================================
# Define my BBs
#==============================================================================
with nd.Cell(name='my_mzm') as mzm:
    eopml = 750
    dx = 150
    dy = dx/2

    # Place building blocks
    mmi_left  = demo.mmi1x2_dp().put()
    eopm_top  = demo.eopm_dc(eopml).put('a0', mmi_left.pin['b0'].move(dx,dy))
    eopm_bot  = demo.eopm_dc(eopml).put('b0', mmi_left.pin['b0'].move(dx,-dy))
    mmi_right = demo.mmi2x2_dp().put('b0', eopm_bot.pin['a0'].move(dx,dy))

    # Interconnect building blocks
    demo.deep.bend_strt_bend_p2p(mmi_left.pin['b0'], eopm_top.pin['a0']).put()
    demo.deep.bend_strt_bend_p2p(mmi_left.pin['b1'], eopm_bot.pin['b0']).put()
    demo.deep.bend_strt_bend_p2p(mmi_right.pin['b1'], eopm_top.pin['b0']).put()
    demo.deep.bend_strt_bend_p2p(mmi_right.pin['b0'], eopm_bot.pin['a0']).put()

    # Define pins for the mzm cell
    nd.Pin(name='a0', xs='Deep', width=demo.wdeep).put(mmi_left.pin['a0'])
    nd.Pin(name='b0', xs='Deep', width=demo.wdeep).put(mmi_right.pin['a0'])
    nd.Pin(name='b1', xs='Deep', width=demo.wdeep).put(mmi_right.pin['a1'])
    nd.Pin(name='c0', xs='MetalRF', width=demo.wmetdc).put(eopm_top.pin['c0'])
    nd.Pin(name='c1', xs='MetalRF', width=demo.wmetdc).put(eopm_top.pin['c1'])
    nd.Pin(name='c2', xs='MetalRF', width=demo.wmetdc).put(eopm_bot.pin['c0'])
    nd.Pin(name='c3', xs='MetalRF', width=demo.wmetdc).put(eopm_bot.pin['c1'])

    nd.nazca_logo.put(mmi_left.pin['a0'].move(-150, 150, 180))


with nd.Cell(name='my_soa') as soas:
    soal1 = 200
    soal2 = 500

    # Place building  blocks
    wg1 = demo.deep.strt(length=150).put()
    wgt1 = demo.d2s().put()
    demo.shallow.strt(length=50).put()
    soa1 = demo.soa(soal1).put()
    demo.shallow.strt(length=100).put()
    soa2 = demo.soa(soal2).put()
    wg2 = demo.shallow.strt(length=50).put()

    # Define pins for the soas cell
    nd.Pin(name='a0', xs='Deep', width=demo.wdeep).put(wg1.pin['a0'])
    nd.Pin(name='b0', xs='Deep', width=demo.wdeep).put(wg2.pin['b0'])
    nd.Pin(name='c0', xs='MetalDC', width=demo.wmetdc).put(soa1.pin['c0'])
    nd.Pin(name='c1', xs='MetalDC', width=demo.wmetdc).put(soa2.pin['c0'])


#==============================================================================
# Design
#==============================================================================
with nd.Cell(name='project') as project:

    # Initialize Die and Package
    die1 = DIE.cell().put(0)
    pack1 = PACK.cell().put(0)
    # Add IOs

    iodeep = demo.io('angled', bend=True, deep=True)
    ioL002 = iodeep.put(die1.pin['ioL002'])
    ioL010 = iodeep.put(die1.pin['ioL010'])
    ioL024 = iodeep.put(die1.pin['ioL024'])
    ioL029 = demo.io('tapered', deep=True).put(die1.pin['ioL029'])
    ioR002 = demo.io('angled', bend=True, deep=False).put(die1.pin['ioR008'])
    ioR029 = demo.io('simple', deep=False).put(die1.pin['ioR029'])
    # Add metal pads
    RFs = addAllBBs(instance=pack1, BBcell=demo.pad_rf(), pinprefix='rf')
    DCTs = addAllBBs(instance=pack1, BBcell=demo.pad_dc(), pinprefix='dcT')
    DCBs = {i: demo.pad_dc().put('rc', pack1.pin['dcB{:03d}'.format(i)]) for i in [5, 6]}

    # Design: put MZM
    iodeep.put(die1.pin['ioL018'])
    mzm1 = mzm.put()

    # Connect MZM1 to IOs
    demo.deep.ubend_p2p(mzm1.pin['b0'], ioL010.pin['b0']).put()
    demo.deep.ubend_p2p(mzm1.pin['b1'], ioL024.pin['b0']).put()
    # Connect MZM1 to RF pins
    demo.metalrf.strt_bend_strt_p2p(mzm1.pin['c0'], RFs['rf003'].pin['c0']).put()
    demo.metalrf.sbend_p2p(mzm1.pin['c1'].rot(-90), RFs['rf002'].pin['c0'], Lstart=700).put()
    demo.metalrf.sbend_p2p(mzm1.pin['c2'].rot(90), RFs['rf001'].pin['c0'], Lstart=700).put()
    demo.metalrf.strt_bend_strt_p2p(mzm1.pin['c3'], RFs['rf000'].pin['c0']).put()

    # Design: put SOAs
    soas1 = soas.put(ioL002.pin['b0'])
    soas2 = soas.put(ioL029.pin['b0'])
    # Connect SOAs to IO and DC pins
    demo.shallow.sbend_p2p(soas1.pin['b0'], ioR002.pin['b0']).put()
    demo.metaldc.ubend_p2p(soas1.pin['c0'], DCBs[6].pin['c0'], length=200).put()
    demo.metaldc.ubend_p2p(soas1.pin['c1'], DCBs[5].pin['c0'], length=100).put()
    demo.shallow.sbend_p2p(soas2.pin['b0'], ioR029.pin['b0']).put()
    demo.metaldc.sbend_p2p(soas2.pin['c0'], DCTs['dcT001'].pin['c0']).put()
    demo.metaldc.sbend_p2p(soas2.pin['c1'], DCTs['dcT002'].pin['c0']).put()
    demo.metaldc.sbend_p2p(soas2.pin['c1'], DCTs['dcT003'].pin['c0']).put()

#nd.export_plt(clear=False)
nd.export_gds(project, filename='demo_lib.gds')

#demo.exportBBlist2GDS() #print and export available BBs
