# -*- coding: utf-8 -*-
"""
Demofab example-1

To view predefined layers in klayout load:
lyp_demo.lyp
"""
import nazca as nd
import nazca.demofab.pdk as demopdk
import nazca.demopackager.packages as demopackages


#==============================================================================
# Create and initialize die and package objects.
#==============================================================================
DIE = demopdk.DesignArea(name='die', coatingR='HR', coatingL='AR')
PACKAGE = demopackages.Package1(name='package1')
PACKAGE.RFedge = 110

# Create a cell based on DIE and package objects.
# This cell acts as your design template.
with nd.Cell('template') as template:

    # put the die cell and make all its pins available to 'template'
    die = DIE.cell().put(0)
    die.raise_pins()

    # put package cell and add building blocks on the pin or add pin attribites
    pack = PACKAGE.cell().put(0)
    for name, pin in pack.pin.items():
        if 'rf' in name:
            # add RFpad and its pin 'c0' to the cell for external connections
            RFpad = demopdk.bb_io_electrical.pad_rf().put('rc', pin)
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        elif 'dcT' in name:
            nd.Pin(name=name, xs='MetalDC').put(pin)
        elif 'dcB' in name:
            nd.Pin(name=name, xs='MetalDC').put(pin)
        else: # all other pins:
            nd.Pin(name=name, pin=pin).put()


#==============================================================================
# Create a design project cell.
#==============================================================================
with nd.Cell(name='project') as project:

    # Place your cell template
    T = template.put(0)

    # put DC pad(s) at the bottom on a package pin
    dcBot = demopdk.bb_io_electrical.pad_dc().put('rc', T.pin['dcB003'])
    dcTop = demopdk.bb_io_electrical.pad_dc().put('rc', T.pin['dcT003'])

    # put optical IO(s) on a package pin
    ioL038 = demopdk.io(shape='angled', bend=True).put(T.pin['ioL038'])

    # Connect a straight waveguide to the optical IO and add a SOA
    demopdk.bb_connectors.shallow.strt(500).put(ioL038.pin['b0'])
    soa1 = demopdk.bb_amplifiers.soa(length=650).put()

    # Connect the soa to DC metal pads
    demopdk.bb_connectors.metaldc.sbend_p2p(\
        soa1.pin['c0'], dcTop.pin['c0'], Lstart=600).put()
    demopdk.bb_connectors.metaldc.sbend_p2p(\
        soa1.pin['c0'].rot(180), dcBot.pin['c0'], Lstart=500).put()


#==============================================================================
# Export project to GDS
#==============================================================================
nd.export_gds(project, filename='project_example1a.gds')
