# -*- coding: utf-8 -*-
"""
Demofab example-2

Create a separate cell containing the die and package templates.

To view predefined layers in klayout load:
lyp_demo.lyp
"""
import nazca as nd
import nazca.demofab.pdk as demopdk
import nazca.demopackager.packages as demopackages

#==============================================================================
# Create and initialize die and package objects.
# Note that you can exchange attributes between die and package.
#==============================================================================
DIE = demopdk.DesignArea(name='die')
DIE.size(length=2500, height=4000)
DIE.coating('AR', 'HR')

PACKAGE = demopackages.Package1(name='package1')
PACKAGE.RFedge = 110
PACKAGE.DCedge = 25
PACKAGE.die_size(length=DIE.length, height=DIE.height)
PACKAGE.DCpitch = 300
PACKAGE.DCcenter = True
PACKAGE.RFcount = 8
PACKAGE.RFcenter = True


# Create a cell based on die and package objects.
# This cell acts as your design template.
with nd.Cell('template') as template:
    die = DIE.cell().put(0)
    die.raise_pins()
    package = PACKAGE.cell().put(0)

    # put all RF pads and top DC pads on the package pins.
    # Add pins to this template cell for later use
    for name, pin in package.pin.items():
        if 'rf' in name:
            RFpad = demopdk.bb_io_electrical.pad_rf().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        elif 'dcT' in name:
            DCpad = demopdk.bb_io_electrical.pad_dc().put('rc', pin)
            # Add DC pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=DCpad.pin['c0']).put()
        else: nd.Pin(name=name, pin=pin).put()


#==============================================================================
# Create a project cell.
#==============================================================================
with nd.Cell(name='project') as project:
    # Place your cell template
    T = template.put(0)
    # Add a specific DC port at the bottom
    dcBot = demopdk.bb_io_electrical.pad_dc().put('rc', T.pin['dcB003'])
    # Add a specific optical IO
    ioL038 = demopdk.io(shape='angled', bend=True).put(T.pin['ioL038'])

    # Connect straight waveguide, SOA, EOPM
    demopdk.bb_connectors.shallow.strt(500).put(ioL038.pin['b0'])
    soa1 = demopdk.bb_amplifiers.soa(length=650).put()
    demopdk.bb_connectors.shallow.strt(200).put()
    demopdk.bb_transitions.s2d().put()
    demopdk.bb_connectors.deep.strt(200).put()
    eopm1 = demopdk.bb_modulators.eopm_dc(length=250, contacts=1, pads=True).put()

    # Connect the soa and the eopm to the DC and RF metal pads
    demopdk.bb_connectors.metaldc.sbend_p2p(soa1.pin['c0'], T.pin['dcT005'], Lstart=600).put()
    demopdk.bb_connectors.metaldc.sbend_p2p(soa1.pin['c0'].rot(180), dcBot.pin['c0'], Lstart=500).put()
    demopdk.bb_connectors.metalrf.strt_bend_strt_p2p(eopm1.pin['c0'], T.pin['rf008']).put()


#==============================================================================
# Export project to GDS
#==============================================================================
nd.export_gds(project, filename='project_example2.gds')
