# -*- coding: utf-8 -*-
"""
Demofab example-3

Add MZM building blocks

To view predefined layers in klayout load:
lyp_demo.lyp
"""
import nazca as nd
import nazca.demofab.pdk as demopdk
import nazca.demopackager.packages as demopackages


#==============================================================================
# Create and initialize die and package objects.
# Note that you can exchange attributes between die and package.
#==============================================================================
DIE = demopdk.DesignArea(name='die')
DIE.size(length=2000, height=4000)
DIE.coating('AR', 'HR')

PACKAGE = demopackages.Package1(name='package1')
PACKAGE.RFedge = 110
PACKAGE.die_size(length=DIE.length, height=DIE.height)
PACKAGE.DCpitch = 300
PACKAGE.DCcenter = True
PACKAGE.RFcount = 4
PACKAGE.RFpitch = 500
PACKAGE.RFcenter = True
PACKAGE.fiberarea = 2000


# Create a cell based on die and package objects.
# This cell acts as your design template.
with nd.Cell('template') as template:
    die = DIE.cell().put(0)
    die.raise_pins()
    package = PACKAGE.cell().put(0)

    # Add all RF pads to the cell
    for name, pin in package.pin.items():
        if 'rf' in name:
            RFpad = demopdk.bb_io_electrical.pad_rf().put('rc', pin)
            # Add RF pin 'c0' to this cell for external connections
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        else: nd.Pin(name=name, pin=pin).put()


#==============================================================================
# Create a MZM cell based on two 1x2 MMIs and two EOPMs.
#==============================================================================
with nd.Cell(name='my_mzm') as mzm:
    eopml = 750
    dx = 150
    dy = dx/2

    # Place building blocks
    mmi_left  = demopdk.bb_splitters.mmi1x2_dp().put()
    eopm_top  = demopdk.bb_modulators.eopm_dc(eopml).put('a0', mmi_left.pin['b0'].move(dx, dy))
    eopm_bot  = demopdk.bb_modulators.eopm_dc(eopml).put('b0', mmi_left.pin['b0'].move(dx, -dy))
    mmi_right = demopdk.bb_splitters.mmi1x2_dp().put('b0', eopm_bot.pin['a0'].move(dx, dy))

    # Interconnect building blocks
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_left.pin['b0'], eopm_top.pin['a0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_left.pin['b1'], eopm_bot.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_right.pin['b1'], eopm_top.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_right.pin['b0'], eopm_bot.pin['a0']).put()

    # Define pins for the mzm cell
    nd.Pin(name='a0', pin=mmi_left.pin['a0']).put()
    nd.Pin(name='b0', pin=mmi_right.pin['a0']).put()
    nd.Pin(name='c0', pin=eopm_top.pin['c0']).put()
    nd.Pin(name='c1', pin=eopm_top.pin['c1']).put()
    nd.Pin(name='c2', pin=eopm_bot.pin['c0']).put()
    nd.Pin(name='c3', pin=eopm_bot.pin['c1']).put()


#==============================================================================
# Create a project cell.
#==============================================================================
with nd.Cell(name='project') as project:
    # Place your cell template
    T = template.put(0)

    # Place MZM
    mzm1 = mzm.put(600, 1300, 90)

    # Add specific optical IOs
    ioL020 = demopdk.io(shape='angled', bend=True, deep=True).put(T.pin['ioL020'])
    ioL056 = demopdk.io(shape='angled', bend=True, deep=True).put(T.pin['ioL056'])

    # Connect the mzm to the optical IOs and to the RF ports
    demopdk.bb_connectors.deep.strt_bend_strt_p2p(mzm1.pin['a0'], ioL020.pin['b0']).put()
    demopdk.bb_connectors.deep.strt_bend_strt_p2p(mzm1.pin['b0'], ioL056.pin['b0']).put()
    demopdk.bb_connectors.metalrf.strt_bend_strt_p2p(mzm1.pin['c0'].rot(90), T.pin['rf000']).put()
    demopdk.bb_connectors.metalrf.strt_bend_strt_p2p(mzm1.pin['c1'].rot(-90), T.pin['rf003']).put()
    demopdk.bb_connectors.metalrf.sbend_p2p(mzm1.pin['c2'], T.pin['rf002'], Lstart=100).put()
    demopdk.bb_connectors.metalrf.sbend_p2p(mzm1.pin['c3'], T.pin['rf001'], Lstart=100).put()

    p = mzm1.pin['a0']

    demopdk.xsShallow.showGuide2D()
#==============================================================================
# Export project to GDS
#==============================================================================
nd.export_gds(project, filename='project_example3.gds')