# -*- coding: utf-8 -*-
"""
Demofab example-4

Create a more complex design with AWGs and MZM.
Create multiple copies of a design project.

To view predefined layers in klayout load:
lyp_demo.lyp
"""
import nazca as nd
import nazca.demofab.pdk as demopdk
import nazca.demopackager.packages as demopackages


#==============================================================================
# Create and initialize die and package objects.
# Note that you can exchange attributes between die and package.
#==============================================================================
DIE = demopdk.DesignArea(name='die')
DIE.size(length=2000, height=4000)
DIE.coating('AR', 'HR')

PACKAGE = demopackages.Package1(name='package1')
PACKAGE.RFedge = 110
PACKAGE.die_size(length=DIE.length, height=DIE.height)
PACKAGE.DCpitch = 300
PACKAGE.DCcenter = True
PACKAGE.RFcount = 4
PACKAGE.RFpitch = 500
PACKAGE.RFcenter = True
PACKAGE.fiberarea = 3000


# Create a cell based on die and package objects.
# This cell acts as your design template.
with nd.Cell('template') as template:
    die = DIE.cell().put(0)
    die.raise_pins()
    package = PACKAGE.cell().put(0)

    # Add all RF pads to the cell
    for name, pin in package.pin.items():
        if 'rf' in name:
            RFpad = demopdk.bb_io_electrical.pad_rf().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        else: nd.Pin(name=name, pin=pin).put()


#==============================================================================
# Create a MZM cell based on two 1x2 MMIs and two EOPMs.
#==============================================================================
with nd.Cell(name='MZM') as mzm:
    eopml = 750
    dx = 150
    dy = dx/2

    # Place foundry building blocks
    mmi_left  = demopdk.bb_splitters.mmi1x2_dp().put(0)
    eopm_top  = demopdk.bb_modulators.eopm_dc(eopml).put('a0', mmi_left.pin['b0'].move(dx, dy))
    eopm_bot  = demopdk.bb_modulators.eopm_dc(eopml).put('b0', mmi_left.pin['b0'].move(dx, -dy))
    mmi_right = demopdk.bb_splitters.mmi1x2_dp().put('b0', eopm_bot.pin['a0'].move(dx, dy))

    # Connect the building blocks point-to-point
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_left.pin['b0'], eopm_top.pin['a0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_left.pin['b1'], eopm_bot.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_right.pin['b1'], eopm_top.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(mmi_right.pin['b0'], eopm_bot.pin['a0']).put()

    # Define pins for the mzm cell
    nd.Pin(name='a0', pin=mmi_left.pin['a0']).put()
    nd.Pin(name='b0', pin=mmi_right.pin['a0']).put()
    nd.Pin(name='c0', pin=eopm_top.pin['c0']).put()
    nd.Pin(name='c1', pin=eopm_top.pin['c1']).put()
    nd.Pin(name='c2', pin=eopm_bot.pin['c0']).put()
    nd.Pin(name='c3', pin=eopm_bot.pin['c1']).put()


#==============================================================================
# Create a project cell.
#==============================================================================
with nd.Cell(name='project') as project:
    # Place your cell template
    T = template.put(0)

    # Place a MZM
    mzm1 = mzm.put(600, 1300, 90)

    # Connect the MZM to the RF ports in T
    demopdk.bb_connectors.metalrf.strt_bend_strt_p2p(mzm1.pin['c0'].rot(90), T.pin['rf000']).put()
    demopdk.bb_connectors.metalrf.strt_bend_strt_p2p(mzm1.pin['c1'].rot(-90), T.pin['rf003']).put()
    demopdk.bb_connectors.metalrf.sbend_p2p(mzm1.pin['c2'], T.pin['rf002'], Lstart=100).put()
    demopdk.bb_connectors.metalrf.sbend_p2p(mzm1.pin['c3'], T.pin['rf001'], Lstart=100).put()

    # Place two AWGs
    awg1 = demopdk.bb_awgs.awg1x4.put(700, 1000, 0)
    awg2 = demopdk.bb_awgs.awg1x4.put(700, 2900, 0, flip=True)

    # Connect the AWGs to the MZM
    demopdk.bb_connectors.deep.strt_bend_strt_p2p(awg1.pin['a0'], mzm1.pin['a0']).put()
    demopdk.bb_connectors.deep.strt_bend_strt_p2p(awg2.pin['a0'], mzm1.pin['b0']).put()

    # Add optical IOs
    io = demopdk.io(shape='angled', bend=True, deep=True)

    ioL010 = io.put(T.pin['ioL010'])
    ioL012 = io.put(T.pin['ioL012'])
    ioL014 = io.put(T.pin['ioL014'])
    ioL016 = io.put(T.pin['ioL016'])

    ioL058 = io.put(T.pin['ioL058'])
    ioL060 = io.put(T.pin['ioL060'])
    ioL062 = io.put(T.pin['ioL062'])
    ioL064 = io.put(T.pin['ioL064'])

    # Connect awg1 to the optical IOs
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg1.pin['b0'], ioL010.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg1.pin['b1'], ioL012.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg1.pin['b2'], ioL014.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg1.pin['b3'], ioL016.pin['b0']).put()
    # Connect awg2 to the optical IOs
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg2.pin['b0'], ioL064.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg2.pin['b1'], ioL062.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg2.pin['b2'], ioL060.pin['b0']).put()
    demopdk.bb_connectors.deep.bend_strt_bend_p2p(awg2.pin['b3'], ioL058.pin['b0']).put()


#==============================================================================
# Place the project multiple times in the top cell
#==============================================================================
project.put(0)
project.put(2*DIE.length-DIE.cleave, 0, 180, flip=True)

#==============================================================================
# Export to GDS
#==============================================================================
nd.export_gds(filename='project_example4.gds')
