# -*- coding: utf-8 -*-
"""
Demofab example-4

Alternative scripting solutions.

Combining multiple projects using bounding box pins on the die outline.

To view predefined layers in klayout load:
lyp_demo.lyp
"""
import nazca as nd
import nazca.demofab.pdk as demopdk
import nazca.demopackager.packages as demopackages


# create some aliases for shorter code:
mmi     = demopdk.bb_splitters.mmi1x2_dp()
eopm750 = demopdk.bb_modulators.eopm_dc(length=750)
deep    = demopdk.bb_connectors.deep
metalrf = demopdk.bb_connectors.metalrf


#==============================================================================
# Create and initialize die and package objects.
# Note that you can exchange attributes between die and package.
#==============================================================================
DIE = demopdk.DesignArea(name='die')
DIE.size(length=2000, height=4000)
DIE.coating('AR', 'HR')

PACKAGE = demopackages.Package1(name='package1')
PACKAGE.die_size(length=DIE.length, height=DIE.height)
PACKAGE.DCside = 30
PACKAGE.RFedge = 110
PACKAGE.DCpitch = 300
PACKAGE.DCcenter = True
PACKAGE.RFcount = 4
PACKAGE.RFpitch = 500
PACKAGE.RFcenter = True
PACKAGE.fiberarea = 3000


# Create a cell based on die and package objects.
# This cell acts as your design template.
with nd.Cell('template') as template:
    die = DIE.cell().put(0)
    die.raise_pins()
    package = PACKAGE.cell().put(0)

    # Add all RF pads to the cell
    for name, pin in package.pin.items():
        if 'rf' in name:
            RFpad = demopdk.pad_rf().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        else: nd.Pin(name=name, pin=pin).put()


#==============================================================================
# Create a MZM cell based on two 1x2 MMIs and two EOPMs.
#==============================================================================
with nd.Cell(name='MZM') as mzm:
    dx = 150
    dy = dx/2

    # Place foundry building blocks
    mmi_left  = mmi.put(0)
    eopm_top  = eopm750.put('a0', mmi_left.pin['b0'].move(dx, dy))
    eopm_bot  = eopm750.put('b0', mmi_left.pin['b0'].move(dx, -dy))
    mmi_right = mmi.put('b0', eopm_bot.pin['a0'].move(dx, dy))

    # Connect the building blocks point-to-point
    deep.bend_strt_bend_p2p(mmi_left.pin['b0'], eopm_top.pin['a0']).put()
    deep.bend_strt_bend_p2p(mmi_left.pin['b1'], eopm_bot.pin['b0']).put()
    deep.bend_strt_bend_p2p(mmi_right.pin['b1'], eopm_top.pin['b0']).put()
    deep.bend_strt_bend_p2p(mmi_right.pin['b0'], eopm_bot.pin['a0']).put()

    # Define pins for the mzm cell
    nd.Pin(name='a0', pin=mmi_left.pin['a0']).put()
    nd.Pin(name='b0', pin=mmi_right.pin['a0']).put()
    nd.Pin(name='c0', pin=eopm_top.pin['c0']).put()
    nd.Pin(name='c1', pin=eopm_top.pin['c1']).put()
    nd.Pin(name='c2', pin=eopm_bot.pin['c0']).put()
    nd.Pin(name='c3', pin=eopm_bot.pin['c1']).put()


#==============================================================================
# Create a project cell.
#==============================================================================
with nd.Cell(name='project') as project:
    # Place your cell template
    T = template.put(0)
    #raise die bounding box pins to the project cell.
    T.raise_pins(nd.bbox_pinnames)

    # Place a MZM
    mzm1 = mzm.put(600, 1300, 90)

    # Connect the MZM to the RF ports in T
    metalrf.strt_bend_strt_p2p(mzm1.pin['c0'].rot(90), T.pin['rf000']).put()
    metalrf.strt_bend_strt_p2p(mzm1.pin['c1'].rot(-90), T.pin['rf003']).put()
    metalrf.sbend_p2p(mzm1.pin['c2'], T.pin['rf002'], Lstart=100).put()
    metalrf.sbend_p2p(mzm1.pin['c3'], T.pin['rf001'], Lstart=100).put()

    # Place two AWGs
    awg1 = demopdk.bb_awgs.awg1x4.put(700, 1000, 0)
    awg2 = demopdk.bb_awgs.awg1x4.put(700, 2900, 0, flip=True)

    # Connect the AWGs to the MZM
    deep.strt_bend_strt_p2p(awg1.pin['a0'], mzm1.pin['a0']).put()
    deep.strt_bend_strt_p2p(awg2.pin['a0'], mzm1.pin['b0']).put()

    # Add optical IOs
    iocell = demopdk.io(shape='angled', bend=True, deep=True)
    io1 = ['ioL010', 'ioL012', 'ioL014' , 'ioL016']
    IO1 = {name: iocell.put(T.pin[name]) for name in io1}
    io2 = ['ioL058', 'ioL060', 'ioL062' , 'ioL064']
    IO2 = {name: iocell.put(T.pin[name]) for name in io2}

    # Connect awg1 to the optical IOs
    for i, name in enumerate(io1):
        demopdk.bb_connectors.deep.bend_strt_bend_p2p(\
            awg1.pin['b'+str(i)], IO1[name].pin['b0']).put()
    # Connect awg2 to the optical IOs
    for i, name in enumerate(io2):
        j = len(io2)-1-i
        demopdk.bb_connectors.deep.bend_strt_bend_p2p(\
            awg2.pin['b'+str(j)], IO2[name].pin['b0']).put()


#==============================================================================
# Place the project multiple times in the top cell
# Note that the (0, 0) mask point w.r.t. the die is a choice of the foundry.
#==============================================================================
#relative placement using the die-ouline
cellA = project.put('lb', 0)
cellB = project.put('bl', cellA.pin['tl'])
cellC = project.put('rb', cellA.pin['rb'], flip=True)

cellD = project.put('lb', cellC.pin['lt'])
cellE = project.put('bl', cellB.pin['tl'])
cellF = project.put('br', cellD.pin['tr'], flip=True)

# loop placement:
for i in range(6):
    project.put('lt', i*DIE.length)

#==============================================================================
# Export to GDS
#==============================================================================
nd.export_gds(filename='project_example5.gds')
