# -*- coding: utf-8 -*-
"""
Demofab example-4

Alternative scripting solutions.

Combining multiple projects using bounding box pins on the die outline.

To view predefined layers in klayout load:
lyp_demo.lyp
"""
import nazca as nd
import nazca.demofab.pdk as demopdk
import nazca.demopackager.packages as demopackages


# create some aliases for shorter code:
mmi        = demopdk.bb_splitters.mmi1x2_dp()
#eopm750    = demopdk.bb_modulators.eopm_dc(length=750)
eopm750    = demopdk.eopm_dc(length=750)
deep       = demopdk.bb_connectors.deep
shallow    = demopdk.bb_connectors.shallow
metalrf    = demopdk.bb_connectors.metalrf
metaldc    = demopdk.bb_connectors.metaldc
#nazca_logo = nd.load_gds(
#    filename='nazca_logo.gds',
#    cellname='nazca',
#    newcellname='nazca_logo')


#==============================================================================
# Create and initialize die and package objects.
# Note that you can exchange attributes between die and package.
#==============================================================================
DIE1 = demopdk.DesignArea(name='die')
DIE1.size(length=2000, height=4000)
DIE1.coating('AR', 'HR')

PACKAGE1 = demopackages.Package1(name='package1')
PACKAGE1.die_size(length=DIE1.length, height=DIE1.height)
PACKAGE1.DCside = 30
PACKAGE1.RFedge = 110
PACKAGE1.DCpitch = 300
PACKAGE1.DCcenter = True
PACKAGE1.RFcount = 4
PACKAGE1.RFpitch = 500
PACKAGE1.RFcenter = True
PACKAGE1.fiberarea = 3000

#create a 2nd die with its own package template
DIE2 = demopdk.DesignArea(name='die2')
DIE2.size(length=4000, height=4000)
DIE2.coating('AR', 'AR')

PACKAGE2 = demopackages.Package1(name='package2')
PACKAGE2.DCside = 20
PACKAGE2.die_size(length=DIE2.length, height=DIE2.height)
PACKAGE2.DCpitch = 200
PACKAGE2.DCcenter = True
PACKAGE2.RFcount = 0
PACKAGE2.fiberarea = 3000

# Create a cell based on die and package objects.
with nd.Cell('template1') as template1:
    die = DIE1.cell().put(0)
    die.raise_pins()
    package = PACKAGE1.cell().put(0)

    # Add all RF pads to the cell
    for name, pin in package.pin.items():
        if 'rf' in name:
            RFpad = demopdk.bb_io_electrical.pad_rf().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        elif 'dc' in name:
            DCpad = demopdk.bb_io_electrical.pad_dc().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=DCpad.pin['c0']).put()
        else: nd.Pin(name=name, pin=pin).put()

# Create a cell based on die and package objects.
with nd.Cell('template2') as template2:
    die = DIE2.cell().put(0)
    die.raise_pins()

    package = PACKAGE2.cell().put(0)

    # Add all RF pads to the cell
    for name, pin in package.pin.items():
        if 'rf' in name:
            RFpad = demopdk.bb_io_electrical.pad_rf().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=RFpad.pin['c0']).put()
        elif 'dc' in name:
            DCpad = demopdk.bb_io_electrical.pad_dc().put('rc', pin)
            # Add RF pin 'c0' to the cell for external connections
            nd.Pin(name=name, pin=DCpad.pin['c0']).put()
        else: nd.Pin(name=name, pin=pin).put()


#==============================================================================
# Create a MZM cell based on two 1x2 MMIs and two EOPMs.
#==============================================================================
with nd.Cell(name='MZM') as mzm:
    dx = 150
    dy = dx/2

    # Place foundry building blocks
    mmi_left  = mmi.put(0)
    eopm_top  = eopm750.put('a0', mmi_left.pin['b0'].move(dx, dy))
    eopm_bot  = eopm750.put('b0', mmi_left.pin['b0'].move(dx, -dy))
    mmi_right = mmi.put('b0', eopm_bot.pin['a0'].move(dx, dy))

    # Connect the building blocks point-to-point
    deep.bend_strt_bend_p2p(mmi_left.pin['b0'], eopm_top.pin['a0']).put()
    deep.bend_strt_bend_p2p(mmi_left.pin['b1'], eopm_bot.pin['b0']).put()
    deep.bend_strt_bend_p2p(mmi_right.pin['b1'], eopm_top.pin['b0']).put()
    deep.bend_strt_bend_p2p(mmi_right.pin['b0'], eopm_bot.pin['a0']).put()

    # Define pins for the mzm cell
    nd.Pin(name='a0', pin=mmi_left.pin['a0']).put()
    nd.Pin(name='b0', pin=mmi_right.pin['a0']).put()
    nd.Pin(name='c0', pin=eopm_top.pin['c0']).put()
    nd.Pin(name='c1', pin=eopm_top.pin['c1']).put()
    nd.Pin(name='c2', pin=eopm_bot.pin['c0']).put()
    nd.Pin(name='c3', pin=eopm_bot.pin['c1']).put()


def soamzm(length=1000):
    with nd.Cell(name='SOAMZM'+str(length)) as smzm:
        dx = 150
        dy = dx/2
        eopm = demopdk.eopm_dc(length=length)

        # Place foundry building blocks
        soa1 = demopdk.soa(650).put()
        shallow.strt(400).put()
        demopdk.bb_transitions.s2d().put()
        deep.strt(100).put()

        mmi_left  = mmi.put()
        eopm_top  = eopm.put('a0', mmi_left.pin['b0'].move(dx, dy))
        eopm_bot  = eopm.put('b0', mmi_left.pin['b0'].move(dx, -dy))
        mmi_right = mmi.put('b0', eopm_bot.pin['a0'].move(dx, dy))

        # Connect the building blocks point-to-point
        deep.bend_strt_bend_p2p(mmi_left.pin['b0'], eopm_top.pin['a0']).put()
        deep.bend_strt_bend_p2p(mmi_left.pin['b1'], eopm_bot.pin['b0']).put()
        deep.bend_strt_bend_p2p(mmi_right.pin['b1'], eopm_top.pin['b0']).put()
        deep.bend_strt_bend_p2p(mmi_right.pin['b0'], eopm_bot.pin['a0']).put()

        # Define pins for the mzm cell
        nd.Pin(name='a0', pin=soa1.pin['a0']).put()
        nd.Pin(name='b0', pin=mmi_right.pin['a0']).put()
        nd.Pin(name='c0', pin=eopm_top.pin['c0']).put()
        nd.Pin(name='c1', pin=eopm_top.pin['c1']).put()
        nd.Pin(name='c2', pin=eopm_bot.pin['c0']).put()
        nd.Pin(name='c3', pin=eopm_bot.pin['c1']).put()
        nd.Pin(name='s0', pin=soa1.pin['c0']).put()

    return smzm


#==============================================================================
# PROJECT 1
#==============================================================================
with nd.Cell(name='project') as project:
    # Place your cell template
    T = template1.put(0)
    #raise die bounding box pins to the project cell
    T.raise_pins(nd.bbox_pinnames)

    # Place a MZM
    mzm1 = mzm.put(600, 1300, 90)

    # Connect the MZM to the RF ports in T
    metalrf.strt_bend_strt_p2p(mzm1.pin['c0'].rot(90), T.pin['rf000']).put()
    metalrf.strt_bend_strt_p2p(mzm1.pin['c1'].rot(-90), T.pin['rf003']).put()
    metalrf.sbend_p2p(mzm1.pin['c2'], T.pin['rf002'], Lstart=100).put()
    metalrf.sbend_p2p(mzm1.pin['c3'], T.pin['rf001'], Lstart=100).put()

    # Place two AWGs
    awg1 = demopdk.bb_awgs.awg1x4.put(700, 1000, 0)
    awg2 = demopdk.bb_awgs.awg1x4.put(700, 2900, 0, flip=True)

    # Connect the AWGs to the MZM
    deep.strt_bend_strt_p2p(awg1.pin['a0'], mzm1.pin['a0']).put()
    deep.strt_bend_strt_p2p(awg2.pin['a0'], mzm1.pin['b0']).put()

    # Add optical IOs
    iocell = demopdk.io(shape='angled', bend=True, deep=True)
    io1 = ['ioL010', 'ioL012', 'ioL014' , 'ioL016']
    IO1 = {name: iocell.put(T.pin[name]) for name in io1}
    io2 = ['ioL058', 'ioL060', 'ioL062' , 'ioL064']
    IO2 = {name: iocell.put(T.pin[name]) for name in io2}

    # Connect awg1 to the optical IOs
    for i, name in enumerate(io1):
        deep.bend_strt_bend_p2p(\
            awg1.pin['b'+str(i)], IO1[name].pin['b0']).put()
    # Connect awg2 to the optical IOs
    for i, name in enumerate(io2):
        j = len(io2)-1-i
        deep.bend_strt_bend_p2p(\
            awg2.pin['b'+str(j)], IO2[name].pin['b0']).put()

    metaldc.ubend_p2p(T.pin['dcB000'], T.pin['dcB004'], length=100).put()
    nd.nazca_logo.put(200, 2000, 0)


# =============================================================================
# PROJECT 2
# =============================================================================
with nd.Cell(name='project2') as project2:

    # Place your cell template
    T = template2.put(0)
    #raise die bounding box pins to the project cell.
    T.raise_pins(nd.bbox_pinnames)

    ioangled = demopdk.io(shape='angled', bend=True, deep=False)
    ioangled2 = demopdk.io(shape='angled', bend=True, deep=True)

    smzm1 = []
    for i in range(4):
        N1 = str(i*16+12)
        N2 = str(i*16+12+6)
        length = 750+i*100

        ioangled.put(T.pin['ioL0'+N1])
        shallow.strt(250).put()
        smzm1.append(soamzm(length).put())

        ioangled2.put(T.pin['ioR0'+N2])
        deep.sbend_p2p(nd.cp.here(), smzm1[-1].pin['b0'], Lstart=200).put()

    soaconn = {0: ['dcB003', 180], 1: ['dcB000', 180],
               2: ['dcT000', 0], 3: ['dcT003', 0]}

    for i, name in soaconn.items():
        metaldc.sbend_p2p(smzm1[i].pin['s0'].rot(name[1]),
            T.pin[name[0]], radius=50, width=100).put()

    modconn = {0: ['dcB007', 'dcB012', 'dcB011', 'dcB008', 180],
               1: ['dcB005', 'dcB014', 'dcB013', 'dcB006', 180],
               2: ['dcT006', 'dcT015', 'dcT016', 'dcT005', 0],
               3: ['dcT008', 'dcT013', 'dcT014', 'dcT007', 0]}

    for i, name in modconn.items():
        metalrf.sbend_p2p(smzm1[i].pin['c0'].rot(name[4]), T.pin[name[0]]).put()
        metalrf.sbend_p2p(smzm1[i].pin['c1'].rot(name[4]), T.pin[name[1]]).put()
        metalrf.sbend_p2p(smzm1[i].pin['c2'].rot(180+name[4]), T.pin[name[2]]).put()
        metalrf.sbend_p2p(smzm1[i].pin['c3'].rot(180+name[4]), T.pin[name[3]]).put()

    nd.nazca_logo.put(200, 2000, 0)


#==============================================================================
# Place the project multiple times in the top cell
# Note: the (0, 0) mask coordinate w.r.t. the die is a choice of the foundry.
#       it may or may not lie on the die's edge.
#==============================================================================
#relative die placement using the die bbox pins
cellA = project.put('lb', 0)
cellB = project.put('rb', cellA.pin['rb'], flip=True)
cellC = project2.put('lt', 0)
cellD = project2.put('lb', cellB.pin['lb'])
cellE = project.put('lb', cellC.pin['rb'])
cellF = project.put('rb', cellE.pin['rb'], flip=True)


#==============================================================================
# Export to GDS
#==============================================================================
#nd.export_plt()
nd.export_gds(filename='project_example6.gds', flat=False)
