#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#
#@Authors: Katarzyna Lawniczuk, Ronald Broeke
#@email: nazca@brightphotonics.eu
#2017(c)

"""
**DEMO PDK building block definition.**

This module is intended to define the **building blocks** of a technology.
Building block definitions makes use of building block template functions.

This module also defines the default **Interconnect objects**,
which greatly facilitate making waveguide or metal interconnects between
building blocks.

To see the available building block and there details
"""


import os

import nazca as nd
import nazca.pdk_template as pdk
import nazca.interconnects as interconnect
import nazca.geometries as geom
import nazca.cfg as cfg

dir_path = os.path.dirname(os.path.abspath(__file__))


#==============================================================================
# Create groups to provide a structure to put building blocks (BB) under.
# This is optional, though useful in navigation BBs.
#==============================================================================

#: Building Block (bb) groups to help find components:
bb_connectors     = pdk.Functional_group() #: inter-connectors
bb_transitions    = pdk.Functional_group() #: xsection transistions
bb_splitters      = pdk.Functional_group() #: power splitter and combiners
bb_combiners      = pdk.Functional_group() #: power splitter and combiners
bb_amplifiers     = pdk.Functional_group() #: SOAs
bb_modulators     = pdk.Functional_group() #: modulators
bb_photodiodes    = pdk.Functional_group() #: photo diodes
bb_mode_adapters  = pdk.Functional_group() #: spot size convertors
bb_io_electrical  = pdk.Functional_group() #: electrical IO (pads)
bb_io_optical     = pdk.Functional_group() #: optical IO (edge or surface)
bb_mode_filters   = pdk.Functional_group() #: mode filters
bb_awgs           = pdk.Functional_group() #: AWG multiplexers
bb_tapers         = pdk.Functional_group() #: tapers between different widths


#==============================================================================
# Get xsection settings for use in this file.
#==============================================================================
xsShallow = nd.get_xsection('Shallow') #: 'Shallow' xsection
xsDeep    = nd.get_xsection('Deep')    #: 'Deep' xsection
xsMetalDC = nd.get_xsection('MetalDC') #: 'MetalDC' xsection
xsMetalRF = nd.get_xsection('MetalRF') #: 'MetalRF' xsection


#==============================================================================
# Get xsection parameter settings for use in this file.
#==============================================================================
wshal  = xsShallow.width
Rshal  = xsShallow.radius
wdeep  = xsDeep.width
Rdeep  = xsDeep.radius
wmetdc = xsMetalDC.width
Rmetdc = xsMetalRF.radius
wmetrf = xsMetalDC.width
Rmetrf = xsMetalRF.radius


#==============================================================================
# Define Interconnects (settings default to values in the xsection attributes).
#==============================================================================
bb_connectors.shallow = interconnect.Interconnect(xs='Shallow') #: 'Shallow' interconnects
bb_connectors.deep    = interconnect.Interconnect(xs='Deep')    #: 'Deep' interconnects
bb_connectors.metaldc = interconnect.Interconnect(xs='MetalDC') #: 'MetalDC' interconnects
bb_connectors.metalrf = interconnect.Interconnect(xs='MetalRF') #: 'MetalRF' interconnects


#==============================================================================
# waveguide transitions
#==============================================================================
bb_transitions.d2s = pdk.Tp_XStransition(
    name='d2s',
    length=75,
    width=55,
    pinwidth={'a0':wdeep, 'b0':wshal},
    xs={'a0':'Deep', 'b0':'Shallow'},
    icon=geom.Tp_xsection_transition(layer1=3, layer2=1)
)

bb_transitions.s2d = pdk.Tp_XStransition(
    name='s2d',
    length=75,
    width=55,
    pinwidth={'a0':wshal, 'b0':wdeep},
    xs={'a0':'Shallow', 'b0':'Deep'},
    icon=geom.Tp_xsection_transition(layer1=1, layer2=3)
)


#==============================================================================
# amplifiers
#==============================================================================
bb_amplifiers.soa = pdk.Tp_SOA(
    name = 'soa',
    length=500,
    width=200,
    dy_io = 30,
    pinwidth={'a0':wshal, 'b0':wshal, 'c0':wmetdc},
    xs={'a0':'Shallow', 'b0':'Shallow', 'c0':'MetalDC'},
    xs_bb='BB',
    metallayer='MetalDC',
    drawmetal=True,
    icon=geom.Tp_icon_strt(bufx=0, bufy=10, layer=10))


#==============================================================================
# splitters
#==============================================================================
bb_mode_filters.mmi1x1_sh = pdk.Tp_Modefilter(
    name='modefilter_sh',
    width=55,
    length=80,
    pinwidth={'a0':wshal,'b0':wshal},
    xs={'a0':'Shallow', 'b0':'Shallow'},
    icon=geom.Tp_icon_mmi(Nin=1, Nout=1, w_in=2.0, l_in=2.0,
        w_gap=4.0, l_tot=10, l_buf=4.0, layer=1)
)

bb_splitters.mmi1x2_sh = pdk.Tp_MMI1x2(
    name='mmi1x2_sh',
    width=55,
    length=80,
    pinwidth={'a0':wshal,'b0':wshal, 'b1':wshal},
    xs={'a0':'Shallow', 'b0':'Shallow', 'b1':'Shallow'},
    offset=2,
    icon=geom.Tp_icon_mmi(Nin=1, Nout=2, w_in=2.0, l_in=2.0, w_gap=2.0,
        l_buf=4.0, layer=1)
)

bb_splitters.mmi2x2_sh = pdk.Tp_MMI2x2(
    name = 'mmi2x2_sh',
    width=60,
    length=350,
    pinwidth={'a0':wshal, 'a1':wshal, 'b0':wshal, 'b1':wshal},
    xs={'a0':'Shallow', 'a1':'Shallow', 'b0':'Shallow', 'b1':'Shallow'},
    offset=4,
    icon=geom.Tp_icon_mmi(Nin=2, Nout=2, w_in=2.0, l_in=2.0, w_gap=2.0,
        l_buf=4.0, layer=1)
)

bb_splitters.mmi1x2_dp = pdk.Tp_MMI1x2(
    name='mmi1x2_dp',
    width=55,
    length=80,
    pinwidth={'a0':wdeep, 'b0':wdeep, 'b1':wdeep},
    xs={'a0':'Deep', 'b0':'Deep', 'b1':'Deep'},
    offset=2,
    icon=geom.Tp_icon_mmi(Nin=1, Nout=2, w_in=2.0, l_in=2.0, w_gap=2.0,
        l_buf=4.0, layer=3)
)

bb_splitters.mmi2x2_dp = pdk.Tp_MMI2x2(
    name = 'mmi2x2_dp',
    width=60,
    length=250,
    pinwidth={'a0':wdeep, 'a1':wdeep, 'b0':wdeep, 'b1':wdeep},
    xs={'a0':'Deep', 'a1':'Deep', 'b0':'Deep', 'b1':'Deep'},
    offset=4,
    icon=geom.Tp_icon_mmi(Nin=2, Nout=2, w_in=2.0, l_in=2.0, w_gap=2.0,
        l_buf=4.0, layer=3)
)


#==============================================================================
# modulators
#==============================================================================
_eopm_dc = pdk.Tp_EOPM_DC(
    name='deep.eopm',
    length=750,
    width=55,
    pinwidth={'a0':wdeep, 'b0':wdeep, 'c0':wmetrf},
    xs={'a0':'Deep', 'b0':'Deep', 'c0':'MetalDC'},
    contacts=2,
    icon=geom.Tp_icon_strt(bufx=0, bufy=15, layer=10)
)


#==============================================================================
# photodiodes
#==============================================================================
bb_photodiodes.pd = pdk.Tp_PhotoDetector(
    name='pd_dp',
    length=50,
    width=55,
    pinwidth={'a0':wdeep, 'b0':wdeep, 'c0':wmetdc},
    xs={'a0':'Deep', 'b0':'Deep', 'c0':'MetalDC'},
    icon=geom.Tp_icon_strt(bufx=0, bufy=12.5, layer=10)
)


#==============================================================================
# io_optical
#==============================================================================
bb_io_optical.io = pdk.Tp_IOtrapezoid(
    name = 'io_sh',
    length=100,
    width=4.5,
    angle=-7,
    pinwidth={'a0':wshal, 'b0':wshal},
    xs={'a0':'Shallow', 'b0':'Shallow'})


#==============================================================================
# io_electrical
#==============================================================================
bb_io_electrical.pad_dc = pdk.Tp_DCpad_rectangle(
    name='pad_dc',
    length=150,
    width=150,
    pinwidth={'c0':wmetdc},
    xs={'c0':'MetalDC'})

bb_io_electrical.pad_rf = pdk.Tp_RFpad(
    name='pad_rf',
    length=80,
    width=80,
    pinwidth={'c0':wmetrf},
    xs={'c0':'MetalRF'},
    icon=geom.Tp_icon_strt(bufx=0, bufy=5, layer=11)
)


# =============================================================================
# Composite building blocks
# =============================================================================
@pdk.hashme('EOPM_dc', 'length', 'contacts', 'pads')
def eopm_dc(length=1000, contacts=3, pads=False, sep=10):
    """EOPM with pad option.

    Args:
        length (float): modulator length in um
        contacts (float): number of stubs or pads.
        pads (bool): draw pads if True (default = False)
        sep (float_: separation of pad bbox from eopm stub.

    Returns:
        Cell: eopm element
    """
    with nd.Cell(hashme=True) as C:
        E1 = _eopm_dc(length=length, contacts=contacts).put(0)
        if pads:
            for i in range(contacts):
                pname = 'c'+str(i)
                E2 = pad_dc().put('rc', E1.pin[pname].move(sep))
                metaldc.strt_p2p(pin1=E1.pin[pname], pin2=E2).put()
                E2.raise_pins(['c0'], [pname])
            E1.raise_pins(['a0', 'b0'])
        else:
            E1.raise_pins()
    return C

@pdk.hashme('PD_dc', 'length', 'pads')
def pd_dc(length=100, pads=False, dy=0):
    """"PhotoDiode with pad option.

    Args:
        length (float): PD length in um
        pads (bool): draw pads if True (default = False)
        sep (float_: separation of pad bbox from pin stub.

    Returns:
        Cell: PD element
    """
    with nd.Cell(hashme=True) as C:
        E1 = pd(length=length).put(0)
        if pads:
            E2 = pad_dc().put('rc', E1.pin['c0'].move(dy))
            metaldc.strt_p2p(pin1=E1.pin['c0'], pin2=E2).put()
            E2.raise_pins(['c0'])
            E1.raise_pins(['a0'])
        else:
            E1.raise_pins()
    return C


#==============================================================================
# Standard IOs
#==============================================================================
@pdk.hashme('io', 'shape', 'bend', 'deep')
def io(shape=None, bend=False, deep=False):
    """Create an IO cells that for edge coupling waveguide interfaces.

    Args:
        shape (str): shape of IO. values: None | 'simple' | 'tapered' | 'angled'
        bend (bool): make the IO connection horizontal (True) for connecting to layout
        deep (bool): add a transition to a deeply-etched waveguides (default = False)

    Returns:
        Cell: io element
    """

    options = ['simple', 'tapered', 'angled']
    if shape is None:
        shape = 'simple'
    if shape not in options:
        print("Error: io shape '{}'not recognized. Options are: {}".\
            format(shape, options))

    pin_indent = 0
    length = 100-pin_indent
    width = 5.0

    if shape == 'tapered':
        with nd.Cell(hashme=True) as C:
            bb_connectors.shallow.strt(length, width, arrow=False).put(0)
            iowg = bb_connectors.shallow.taper(length, width, 3.0, arrow=False).put()
            if deep:
                bb_connectors.shallow.strt(10, arrow=False).put()
                iowg = bb_transitions.s2d().put()
            nd.Pin('b0').put(iowg.pin['b0'])

    elif shape == 'angled':
        with nd.Cell(hashme=True) as C:
            if not bend:
                bb_io_optical.io().put(0)
                iowg = bb_connectors.shallow.taper(length, 4.5, 3.0, arrow=False).put()
                if deep:
                    bb_connectors.shallow.strt(10, arrow=False).put()
                    nd.make_arrow().put()
                    iowg = bb_transitions.s2d().put()
                else:
                    nd.make_arrow().put()
                nd.Pin('b0', pin=iowg.pin['b0']).put()

            else:
                bb_io_optical.io().put(0)
                bb_connectors.shallow.taper(length, 4.5, 3.0, arrow=False).put()
                iowg = bb_connectors.shallow.bend(angle=-7, arrow=False).put()
                if deep:
                    bb_connectors.shallow.strt(10, arrow=False).put()
                    nd.make_arrow().put()
                    iowg = bb_transitions.s2d().put()
                else:
                    nd.make_arrow().put()
                nd.Pin('b0', pin=iowg.pin['b0']).put()

    else: #shape == None or shape == 'simple':
        with nd.Cell(hashme=True) as C:
            if not deep:
                iowg = bb_connectors.shallow.strt(length).put(0)
            else:
                bb_connectors.shallow.strt(length+10).put(0)
                iowg = bb_transitions.s2d().put()
            out = nd.Pin('b0', iowg.pin['b0']).put()
            nd.make_arrow().put(out)

    return C


#==============================================================================
# GDS based BBs
#==============================================================================
with nd.Cell('awg1x4') as awg1x4:

    awg = nd.load_gds(
        filename= os.path.join(dir_path, 'gdsBB', 'AWG_1x4.gds'),
        cellname='awg',
        newcellname='awg',
        layermap={1:3})
    awg.put()

    nd.Pin('a0', xs='Deep', width=wdeep).put((9.184851e-016,-5,-90))
    nd.Pin('b0', xs='Deep', width=wdeep).put((273.63575,-4.5798919,-85.569401))
    nd.Pin('b1', xs='Deep', width=wdeep).put((269.89849,-4.9532407,-88.523134))
    nd.Pin('b2', xs='Deep', width=wdeep).put((266.1423,-4.9532407,-91.476866))
    nd.Pin('b3', xs='Deep', width=wdeep).put((262.40504,-4.5798919,-94.430599))
    pdk.put_stub(['a0', 'b0', 'b1', 'b2', 'b3'])

bb_awgs.awg1x4 = awg1x4



#==============================================================================
# Short alias for grouped BBs:
#==============================================================================
shallow   = bb_connectors.shallow
deep      = bb_connectors.deep
metaldc   = bb_connectors.metaldc
metalrf   = bb_connectors.metalrf

soa       = bb_amplifiers.soa
awg1x4    = bb_awgs.awg1x4,
pad_dc    = bb_io_electrical.pad_dc
pad_rf    = bb_io_electrical.pad_rf

bb_modulators.eopm_dc = eopm_dc

pd        = bb_photodiodes.pd

mmi1x1_sh = bb_mode_filters.mmi1x1_sh
mmi1x2_sh = bb_splitters.mmi1x2_sh
mmi2x2_sh = bb_splitters.mmi2x2_sh
mmi1x2_dp = bb_splitters.mmi1x2_dp
mmi2x2_dp = bb_splitters.mmi2x2_dp
s2d       = bb_transitions.s2d
d2s       = bb_transitions.d2s



#==============================================================================
# List of all BBs
#==============================================================================
BBcells = [
    bb_amplifiers.soa(),

    bb_awgs.awg1x4,

    bb_io_electrical.pad_dc(),
    bb_io_electrical.pad_rf(),
    bb_io_optical.io(),

    #bb_modulators.eopm_dc(contacts=0, pads=True),
    #bb_modulators.eopm_dc(contacts=1, pads=True),
    #bb_modulators.eopm_dc(contacts=2, pads=True),

    eopm_dc(pads=False),
    eopm_dc(pads=True, contacts=1),
    eopm_dc(pads=True, contacts=3),

    #bb_photodiodes.pd(pad=False),
    #bb_photodiodes.pd(pad=True),
    pd_dc(pads=True),
    pd_dc(pads=False),

    bb_mode_filters.mmi1x1_sh(),

    bb_splitters.mmi1x2_sh(),
    bb_splitters.mmi2x2_sh(),
    bb_splitters.mmi1x2_dp(),
    bb_splitters.mmi2x2_dp(),

    bb_transitions.s2d(),
    bb_transitions.d2s(),

    bb_connectors.shallow.strt(),
    bb_connectors.shallow.bend(),
    bb_connectors.deep.strt(),
    bb_connectors.deep.bend()
    ]

BBnames = [bb.cell_name for bb in BBcells]

def exportBBlist2GDS():
    """Generate GDS file with all the building blocks."""
    global BBcells
    print('BBs:')
    for bb in BBcells:
        print('  ', bb.cell_name)
    nd.export_gds(BBcells, filename='BB-list.gds')
    return BBcells


def export_gds(topcells=None, filename=None):
    nd.export_gds(topcells=topcells, filename=filename)

