#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#==============================================================================
# (c) 2017 Katarzyna Lawniczuk, Ronald Broeke
#==============================================================================
"""
DEMO Package template
"""

import nazca as nd
import nazca.pdk_template_core as core
import nazca.geometries as geom
import nazca.cfg as cfg

#==============================================================================
# Package template
#==============================================================================
annotationlayer = 'AnnotationIO'
fiberIOlayer = 'FiberIO'
textheight = 0.15
textmove = -0.08
RFname = 'rf{:03d}'
DCtopname = 'dcT{:03d}'
DCbotname = 'dcB{:03d}'


class Package1():
    """
    Class containing a Demo package template.
    """

    def __init__(self, name='Package-Demo',
            length=4000, height=4000, cleave=100,
            DCside=250, DCedge=20, DCpitch=180, DCcount=None, DCcenter=False,
            RFedge=25, RFside=500, RFpitch=200, RFcount=None, RFcenter=False,
            show_fiberarea=True, fiberarea=1000,
            double_row_DC=False, DCx_doublerow=None, DCy_doublerow=None,
            double_row_RF=False, RFx_doublerow=None, RFy_doublerow=None,
            textlayer=None):
        """Construct a Package1 object.

        Args:
            name (str): name of the package cell
            lenght (float): length of the package
            width (float): width of the package
            cleave (float): thickness of cleave/dice area

            DCside (float): distance from side for DC pads
            DCedge (float): distance from the bond-edge for DC pads

            RFside (float): distance from edge for RF pads
            RFedge (float): distance from the bond-edge for RF pads

        Returns :
            None
        """
        self.name   = name
        self.length = length
        self.height = height
        self.cleave = cleave

        self.DCside = DCside
        self.DCedge = DCedge
        self.DCpitch = DCpitch
        self.DCcenter = DCcenter
        self.DCcount = DCcount

        self.RFedge = RFedge
        self.RFside = RFside
        self.RFpitch  = RFpitch
        self.RFcenter = RFcenter
        self.RFcount = RFcount

        self.show_fiberarea = show_fiberarea
        self.fiberarea = fiberarea
        self.double_row_DC = double_row_DC
        self.DCx_doublerow = DCx_doublerow
        self.DCy_doublerow = DCy_doublerow
        self.double_row_RF = double_row_RF
        self.RFx_doublerow = RFx_doublerow
        self.RFy_doublerow = RFy_doublerow
        self.textlayer = textlayer

        if self.textlayer is None:
            self.textlayer = 'AnnotationIO'
        else:
            self.textlayer = textlayer

        self.arrow = core.make_arrow(layer=cfg.package_pin_layer)


    def die_size(self, length=4000, height=4000):
        """Set size of the package area.

        Args:
            length (float): length of the package in um
            heigth (float): height of the package in um

        Returns:
            None
        """
        self.length = length
        self.height = height


    def cell(self):
        """Create a Cell with DC, RF, Optical fiber postitions.

        Returns:
            Cell: package cell
        """
        with nd.Cell(name=self.name) as C:

            #Placing PAD IOs
            IOcountmax = round((self.length-2*self.DCside)/self.DCpitch)
            if self.DCcount is not None and self.DCcount < IOcountmax:
                IOcount = self.DCcount
            else:
                IOcount = IOcountmax

            #DC PAD positions
            for n in range(0, IOcount):
                pinIDT = DCtopname.format(n)
                pinIDB = DCbotname.format(n)

                if self.DCcenter:
                    dist_edgex = self.length - self.cleave -\
                        (IOcount-1)*self.DCpitch-2*self.DCside
                else: dist_edgex = 0

                #Top DC ports
                p = nd.Pin(pinIDT).put(
                    n*self.DCpitch+self.DCside+0.5*dist_edgex,
                    self.height-self.cleave-self.DCedge,
                    -90)
                self.arrow.put(p)
                text = nd.text(pinIDT, layer=annotationlayer,
                    height=textheight, align='rc')
                text.put(p.move(textmove))

                #Bottom DC ports
                p = nd.Pin(pinIDB).put(
                    n*self.DCpitch+self.DCside+0.5*dist_edgex,
                    self.DCedge,
                    90)
                self.arrow.put(p)
                text = nd.text(pinIDB, layer=annotationlayer,
                    height=textheight, align='rc')
                text.put(p.move(textmove))

            #DC PAD positions double row
            if self.double_row_DC:
                for ndr in range(0, IOcount-1):
                    n = ndr+IOcountmax
                    pinIDT = DCtopname.format(n)
                    pinIDB = DCbotname.format(n)

                    if self.DCcenter == True:
                        dist_edgex = self.length - self.cleave -\
                            (IOcount-1)*self.DCpitch-2*self.DCside
                    else: dist_edgex = 0

                    if self.DCx_doublerow == None:
                        self.DCx_doublerow = 0.5*self.DCpitch

                    if self.DCy_doublerow == None:
                        self.DCy_doublerow = 1.0*self.DCpitch

                    #Top DC ports
                    p = nd.Pin(pinIDT).put(
                        ndr*self.DCpitch+self.DCside+0.5*dist_edgex+self.DCx_doublerow,
                        self.height-self.cleave-self.DCedge-self.DCy_doublerow,
                        -90)
                    self.arrow.put(p)
                    text = nd.text(pinIDT, layer=annotationlayer,
                        height=textheight, align='rc')
                    text.put(p.move(textmove))

                    #Bottom DC ports
                    p = nd.Pin(pinIDB).put(
                        ndr*self.DCpitch+self.DCside+0.5*dist_edgex+self.DCx_doublerow,
                        self.DCedge+self.DCy_doublerow,
                        90)
                    self.arrow.put(p)
                    text = nd.text(pinIDB, layer=annotationlayer,
                        height=textheight, align='rc')
                    text.put(p.move(textmove))

            #RF PAD positions
            IOcountmax = round((self.height-2*self.RFside)/self.RFpitch)
            if self.RFcount is not None and self.RFcount < IOcount:
                IOcount = self.RFcount
            else:
                IOcount = IOcountmax

            for n in range(0, IOcount):
                piRFn = RFname.format(n)

                if self.RFcenter:
                    dist_edgey = self.height-self.cleave -\
                        (IOcount-1)*self.RFpitch-2*self.RFside
                else: dist_edgey = 0

                p = nd.Pin(piRFn).put(
                    self.length-self.cleave-self.RFedge,
                    n*self.RFpitch+self.RFside+0.5*dist_edgey,
                    180)
                self.arrow.put(p)
                text = nd.text(piRFn, layer=annotationlayer,
                    height=textheight, align='rc')
                text.put(p.move(textmove))

            #RF PAD positions double row
            if self.double_row_RF:
                for ndr in range(0, IOcount-1):
                    n = ndr+IOcountmax
                    piRFn = RFname.format(n)

                    if self.RFcenter == True:
                        dist_edgey = self.height-self.cleave -\
                            (IOcount-1)*self.RFpitch-2*self.RFside
                    else: dist_edgey = 0

                    if self.RFx_doublerow == None:
                        self.RFx_doublerow = 1.0*self.RFpitch

                    if self.RFy_doublerow == None:
                        self.RFy_doublerow = 0.5*self.RFpitch

                    p = nd.Pin(piRFn).put(\
                        self.length-self.cleave-self.RFedge-self.RFx_doublerow,
                        ndr*self.RFpitch+self.RFside+0.5*dist_edgey+self.RFy_doublerow,
                        180)
                    self.arrow.put(p)
                    text = nd.text(piRFn, layer=annotationlayer,
                        height=textheight, align='rc')
                    text.put(p.move(textmove))

            #Fiber positions
            dist_edgey = self.height-self.cleave-self.fiberarea
            faalen = self.cleave+50

            if self.show_fiberarea:
                farea = geom.box(width=self.fiberarea, length=faalen)
                poly = nd.Polygon(layer=fiberIOlayer, points=farea)
                poly.put(-faalen-0.5*self.cleave, 0.5*dist_edgey+0.5*self.fiberarea, 0)
            else: print('Set show_fiberarea == True to see the fiber allowed area for this package.')

        return C


    @property
    def maxDCcount(self):
        """Get max number of DC ports that fit in a package

        Returns:
            int: maximum number of DC pads
        """
        return round((self.length-2*self.DCside)/self.DCpitch)


    @property
    def maxRFcount(self):
        """Get max number of RF ports that fit in a package.

        Returns:
            int: maximum number of RF pads
        """
        return round((self.length-2*self.RFside)/self.RFpitch)
