#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#==============================================================================
# (c) 2016-2017 Ronald Broeke, Katarzyna Lawniczuk
#==============================================================================

"""Module with a set Buiding Block templates for faciliating PDK creation."""

import os
from collections import OrderedDict
import inspect
from functools import wraps
import hashlib
import pandas as pd
import nazca as nd
from nazca import cfg
from nazca.util import parameters_to_string
import nazca.geometries as geom


#==============================================================================
# hash
#==============================================================================
_hash_id = None
_hash_params = {}
_hash_name = 'NONAME'
cfg._hashme = []

def hashme(*params):
    """Parametrize the decorator to send base name and parameter names.

    Returns:
        decorator
    """
    #print('params:' , params)
    Npar = len(params)
    if Npar == 0:
        print('warning: no name given to @hashme')
        name = 'NONAME'
    else:
        name = params[0]
        params = params[1:]
    def decorator(cellfunc):
        """Decorator for grabbing and hashing a function's full parameter list.

        While executing the wrapper a global hashid and parameter list are set.
        Before leaving the wrapper the hashid is reset to None.

        Returns:
            function: wrapper function
        """
        @wraps(cellfunc)
        def wrapper(*args, **kwargs):
            global _hash_id
            global _hash_params
            global _hash_name
            nonlocal name

            hash_length = 4
            name_long = name
            getargs = inspect.getargspec(cellfunc)
            funcargs = getargs.args
            funcdefaults = getargs.defaults
            hashstr = ''
            _hash_params = OrderedDict()

            if funcargs:
                # also make it work for class methods by removing 'self':
                skip = 0
                if funcargs[0] == 'self':
                    skip = 1
                    funcargs = funcargs[1:]
                    print(
"""Warning: @hashme decorator on Class method:
The @hashme "state checker' can only be used savely if the state of the
function it decorates is *not* dependend on a state *external* to that
function, i.e. if the variations of the cell it generates *only* depend
on the function parameters. In contrast, Class methods typically depend
on attributes stored at class level, hence, by their very nature Class methods
do not fit well with the @hashme concept to guarantee unique cells for
unique names.

By default Nazca will guarantee a unique cell name each time a cell is created.
However, using @hashme returns an existing cell if there is a name match.
A secure way in all cases is call the cell generation function only once
and assign its returns value, the Cell object, to a variable and use the
variable to put instances.

Example:
    class myElement():
        ...
        def make_cell(a, b, c):
            ...
            return cell_object

    elm = myElement()

# method 1:
# a) creates two copies of the cell.
# b) if @hashme is used it returns an existing cell if the call profile
#    of make_cell has been used earlier. This creates one copy of the cell.

    elm.cell(a, b, c).put()
    elm.cell(a, b, c).put()

# method 2:
# reuse the unmutable cell E:

    E = elm.cell(a, b, c)
    E.put()
    E.put()
""")

                for i, (a, v) in enumerate(zip(funcargs, funcdefaults)):
                    try: # set to kwargs
                        _hash_params[a] = kwargs[a]
                    except:
                        try: # set to args
                            if a != 'self':
                                _hash_params[a] = args[i+skip]
                        except: # set to default
                            _hash_params[a] = v

                for p in sorted(_hash_params.keys()):
                    hashstr += '{}_{}'.format(p, _hash_params[p])
                for p in params:
                    if p in _hash_params.keys():
                        name_long += "_{}".format(_hash_params[p])

                _hash_id = hashlib.md5(hashstr.encode()).hexdigest()[:hash_length]
            else:
                _hash_id = ''

            if _hash_id is not '':
                _hash_name = "{}_${}".format(name_long, _hash_id)
            else:
                _hash_name = name_long

            # add cfg version to avoid importing template_core for hashme attributes
            cfg.hash_id = _hash_id
            cfg.hash_params = _hash_params
            cfg.hash_basename = name            # base
            cfg.hash_paramsname = name_long  # base + params
            cfg.hash_name = _hash_name       # base + params + hash

            cell_exists = _hash_name in cfg.cellnames.keys()
            if cell_exists:
                return cfg.cellnames[_hash_name]
            else:
               cell_new = cellfunc(*args, **kwargs)

            #reset:
            _hash_id = None
            _hash_params = {}
            _hash_name = ''
            cfg.hash_id  = ''
            cfg.hash_params = ''
            cfg.hash_basename = ''
            cfg.hash_paramsname = ''
            cfg.hash_name = ''
            return cell_new
        return wrapper
    return decorator


#==============================================================================
#
#==============================================================================
stubs = dict() # All stubs. {stub_name: stub_cell_obj}
stubmap = dict() # Map xs to its stub_xs. {xs_name: stub_xs_name}


def create_template_layers():
    """Define the default mask-layers used in the pdk_template.

    Call this function when the default mask layers are missing.
    Layers content can overwritten by specific foundry layer definitions later.

    Returns:
        None
    """
    nd.add_layer('Annotation', 1112)
    nd.add_layer('AnnotationPin', 1112)
    nd.add_layer('AnnotationIO', 1112)
    nd.add_layer('BBtext', 1112)
    nd.add_layer('BB', 1112)


class Functional_group():
    """
    Class to group building blocks syntactically together.

    e.g. a bb_splitters group with various splitter components:

        bb_splitters.mmi1x2

        bb_splitters.mmi2x2

        bb_splitters.mmi3x3
    """

    def __init__(self, name=''):
        self.name = name
BB_Group = Functional_group


def parameters(parameters=None):
    """Put parameter list as annotation in a building block.

    Args:
        parameters (dict): {parameter_name: value}

    Returns:
        None
    """

    #store parameters as a dict in the cell for PDK checks

    cell = cfg.cells[-1]
    if isinstance(parameters, dict):
        cell.parameters = parameters
        anno = nd.Annotation(text=parameters_to_string(parameters),
            layer='Annotation')
    elif cell.hashme:
        anno = nd.Annotation(text=parameters_to_string(cell.parameters),
            layer='Annotation')
    else:
        anno = nd.Annotation(text='', layer='Annotation')
    return anno


def cellname(cellname=None, length=0, align='lc'):
    """Create text of the cellname and position current pin to place it.

    Args:
        cellname (str): name of the cell
        length (float): length available for the BB name in um
        align (str): text alignment (see nazca.text, default = 'lc')

    Returns:
        Cell: text with cellname formated
    """
    cell = cfg.cells[-1]
    if cellname is None:
       cellname = cell.cell_paramsname
    maxheight = cfg.cellname_max_height
    length = length * cfg.cellname_scaling
    texth = min(maxheight, abs(length) / nd.linelength(cellname, 1))
    txt = nd.text(cellname, texth, layer='BBtext', align=align)
    return txt


def addBBmap(name):
    """Apply BB mapping for OD software.

    Returns:
        None

    Exceptions:
        ValueError
    """
    if not cfg.spt:
        return None

    mapping = cfg.bbmap.loc[cfg.bbmap['Nazca'] == name]

    if mapping.empty:
        raise ValueError("Nazca name '{}' not found in '{}'.".format(name, 'cfg.bbmap'))

    try:
        od_name = mapping['OD_BBname'].iloc[0]
    except:
        raise ValueError("No OD_BBname in '{}' for '{}'.".format('cfg.bbmap', name))
    try:
        od_port = mapping['OD_port'].iloc[0]
    except:
        raise ValueError("No OD_port in '{}' for '{}'.".format('cfg.bbmap', name))
    try:
        x, y, a = mapping[['x', 'y', 'a']].iloc[0]
    except:
        raise ValueError("No (x, y, a) in '{}' for '{}'.".format('cfg.bbmap', name))

    #store mapping-info in cell
    cfg.cells[-1].BBmap = (od_name, od_port, (x, y, a))


#==============================================================================
# Stubs
#==============================================================================
arrow = None

@hashme('arrow', 'layer', 'arrowtype')
def make_arrow(layer=None, arrowtype=None):
    """Create a cell with an arrow shape to indicate pin positions.

    Args:
        layer (float): layer number to place the arrow
        arrowtype (str): geometry type of the arrow: full | lefthalf | righthalf

    Returns:
        None
    """
    #Call after foundry layers have been loaded.
    if layer is None:
        layer = 'AnnotationPin'

    if layer not in cfg.layerdict.keys():
        try:
           layer = cfg.default_layer_Annotation
           nd.add_layer(name=layer, layer=layer)
        except:
            pass

    size = cfg.arrowsize
    if arrowtype in ['full', 'lefthalf', 'righthalf']:
        size *= 0.5

    if arrowtype is 'righthalf':
        with nd.Cell(_hash_name) as arrow:
            outline = [(0, 0), (-0.4, -0.4), (-0.4, -0.20), (-0.6, -0.20), (-0.6, 0.0)]
            outline = [(x*size, y*size) for x, y in outline]
            nd.Polygon(layer=layer, points=outline).put(0)
        return arrow

    if arrowtype is 'lefthalf':
        with nd.Cell(_hash_name) as arrow:
            outline = [(0, 0), (-0.4, 0.4), (-0.4, 0.20), (-0.6, 0.20), (-0.6, 0.0)]
            outline = [(x*size, y*size) for x, y in outline]
            nd.Polygon(layer=layer, points=outline).put(0)
        return arrow

    with nd.Cell(_hash_name) as arrow:
        outline = [(0, 0), (-0.4, -0.4), (-0.4, -0.20), (-0.6, -0.20),
            (-0.6, 0.20), (-0.4, 0.20), (-0.4, 0.4)]
        outline = [(x*size,y*size) for x, y in outline]
        nd.Polygon(layer=layer, points=outline).put(0)
    return arrow


def stubname(xs, width, thick, shape=None):
    """Construct a stub name.

    Args:
        xs (str); xsection name
        width (float): stub width
        thick (float): thickness of stub into cell (length)

    Returns:
        str: stub name
    """
    if width is None:
        return 'stub_{}'.format(xs)
    else:
        return 'stub_{}_w{}_t{}_{}'.format(xs, width, thick, shape)


def makestub(xs_guide, width, length=2.0, shape=None, arrowtype=None):
    """Factory for creating stubs in the logical layers.

    Args:
        xs_guide (str): name of xsection
        width (float): stub width
        thick (float): thickness of stub into cell (length)
        shape (str): shape of the stub: circ | rect (default = 'rect')

    Returns:
        None
    """
    global stubs, arrow

    arrow = make_arrow(arrowtype=arrowtype)
    try:
        xs_logic = cfg.stubmap[xs_guide]
    except:
        # if no stub defined, use the xs as its own stub.
        #if xs_guide is None:
        #    xs_guide = 'None'
        if xs_guide is None:
            return arrow

        if xs_guide not in cfg.XSdict.keys():
            raise Exception("Can not make a stub in an undefined xsection '{0}'.\n"\
               "Possible causes: '{0}' is misspelled or not yet defined.\n"\
               "To define the xsection:\n"\
               "add_xsection(name='{0}')\n"\
               "or with layers info and adding a custom stub:\n"\
               "add_xsection(name='{0}', layer=1)\n"\
               "add_xsection(name='{1}', layer=2)\n"\
               "add_stub(xsection='{0}', stub='{1}')".format(xs_guide, 'stubname'))

        cfg.stubmap[xs_guide] = xs_guide
        xs_logic = xs_guide


    shapes = ['box', 'circle']
    name = stubname(xs_guide, width, length, shape)
    if width is None:
        width = 0
    with nd.Cell(name) as C:
        arrow.put(0)

        for lay, growx, acc in nd.layeriter(xs_logic):
            if shape is 'circle':
                outline = geom.circle(radius=0.5*length, N=32)
            else:
                outline = geom.box(width=width+2*growx, length=length)
            nd.Polygon(layer=lay, points=outline).put(0, 0, 180)

            if shape not in shapes:
                print("Warning: stub shape '{}' not recognized, possible options are {}.".
                    format(shape, shapes))
    stubs[name] = C
    return C


def put_stub(pinname=None, length=2.0, shape='box', arrowtype=None, layer=None):
    """Add a xsection stub to a pin.

    The xsection and width of the stub is obtained from the pin attributes.

    Args:
        pinname (str | list of str): name(s) of the pin(s) (default = None)
            If pinname is None all the chained pins will obtain a stub.
        length (float): length of the stub (thickness into cell)

    Returns:
        None
    """

    if isinstance(pinname, str):
        pinname = [pinname]

    if isinstance(pinname, nd.Node):
        pinname = [pinname.name]

    if pinname is None or pinname == []:
        #raise ValueError('pinname requries a string name of the pin or list of string names.')
        pinname = [name for name, pin in cfg.cells[-1].pin.items() if pin.chain == 1]

    for p in pinname:
        cell = cfg.cells[-1]
        try:
            node = cell.pin[p]
            if node.xs is None:
                xs = None
                width = 0
            else:
                xs = node.xs
                width = node.width

            node.show = True #display pin name in layout
            if xs is None:
                arrow = make_arrow(arrowtype=arrowtype, layer=layer)
                arrow.put(node)
            else:
                name = stubname(xs, width, length, shape)
                if name not in stubs.keys():
                    makestub(xs_guide=xs, width=width, length=length,
                        shape=shape, arrowtype=arrowtype)
                stubs[name].put(node)

                # Move the annotation a bit so it is inside the arrow and the labels of
                # connected pins can be distinguished.
                #TODO: 0.3 should be parametrized.
            nd.Annotation(layer='AnnotationPin', text=p).put(node.move(-0.3))
        except:
            print("warning (core): can't add stub in cell '{}': pin '{}' not existing".\
                format(cell.cell_name, p))


def bb_fingerprint(bb_cells, export_png=False, path='', bbox_stubs=True):
    """Generate a dict with parameter list and pin list of a list off BBs.

    Args:
        bb_cells (list of Cell): all cells to loop over
        export_png (bool): generate png output and save in <path>
        path (str): path for saveing png files
        bbox_stubs (bool): add stubs to the bbox (default = True).
            Only relevant with export_png True

    Returns:
        dict: Dictionary with building block info, Structure is as follows::

            {'<cellname>':
              {'pins':
                { '<pinname>':
                  {'width': <value>,
                   'xs'   : <value>
                   'xya'  : (<x>, <y>, <a>)
                   'show' : <values>
                  }
                  , ...
                },
               'parameters':
                 {'<keywword>': <value>}
                 , ...
              }
            }
    """
    BBsettings = {}
    for bb in bb_cells:
        try:
            groupname = bb.groupname
        except:
            groupname = ''

        if export_png:
            bbox_stubs_store = cfg.bbox_stubs
            cfg.bbox_stubs = bbox_stubs
            nd.export_plt(bb, output='file', path=path, title=bb.cell_paramsname)
            cfg.bbox_stubs = bbox_stubs_store

        pinsetting = {}
        for name, pin in sorted(bb.pin.items()):
            if name != 'org':
                pinsetting[name] = {
                    'width':pin.width,
                    'xsection':pin.xs,
                    'pointer':pin.xya(),
                    'show':pin.show,
                    'type':pin.type
                }

        paramsetting = {}
        try: # parameters may be empty
            for p in bb.parameters:
                paramsetting[p] = bb.parameters[p]
        except:
            pass
        BBsettings[bb.cell_name] = {
            'pins': pinsetting,
            'parameters': paramsetting,
            'paramsname': bb.cell_paramsname,
            'basename': bb.cell_basename,
            'groupname': groupname}

    return BBsettings



bbox_pinnames = [
    'lb', 'lc', 'lt',
    'tl', 'tc', 'tr',
    'rt', 'rc', 'rb',
    'br', 'bc', 'bl',
    'cc']

def put_boundingbox(pinname, length, width, raise_pins=True, outline=True,
        align='lc', name=True, params=True, move=(0, 0, 0)):
    """Create  bounding box (bbox) cell.

    Args:
        pin (str): pin to place bbox on (center left of bbox)
        length (float): length of the bbox
        with (float): width of the bbox
        raise_pins (bool): raise bbox pins into active cell (default = True)
        outline (bool): draw bbox outline (default = True)
        align (str): align the bbox on pin <pinname> (default = 'lc')

    Return:
        None
    """
    _paramsname = cfg.cells[-1].cell_paramsname
    _parameters = cfg.cells[-1].parameters
    with nd.Cell(name='bbox', instantiate=False) as bbox:
        outline = geom.box(length, width)
        nd.Polygon(layer='BB', points=outline).put(0)

        stbs = cfg.bbox_stubs
        nd.Pin('lb', type='bbox', show=stbs).put(0, -0.5*width, 180)
        nd.Pin('lc', type='bbox', show=stbs).put(0, 0, 180)
        nd.Pin('lt', type='bbox', show=stbs).put(0, 0.5*width, 180)

        nd.Pin('tl', type='bbox', show=stbs).put(0, 0.5*width, 90)
        nd.Pin('tc', type='bbox', show=stbs).put(0.5*length, 0.5*width, 90)
        nd.Pin('tr', type='bbox', show=stbs).put(length, 0.5*width, 90)

        nd.Pin('rt', type='bbox', show=stbs).put(length, 0.5*width, 0)
        nd.Pin('rc', type='bbox', show=stbs).put(length, 0, 0)
        nd.Pin('rb', type='bbox', show=stbs).put(length, -0.5*width, 0)

        nd.Pin('br', type='bbox', show=stbs).put(length, -0.5*width, -90)
        nd.Pin('bc', type='bbox', show=stbs).put(0.5*length, -0.5*width, -90)
        nd.Pin('bl', type='bbox', show=stbs).put(0, -0.5*width, -90)

        nd.Pin('cc', type='bbox', show=cfg.bbox_stubs).put(0.5*length, 0, 0)

        if stbs:
            nd.put_stub(['lb', 'tl', 'rt', 'br'], arrowtype='righthalf',
                layer=cfg.bbox_pin_layer)
            nd.put_stub(['lt', 'tr', 'rb', 'bl'], arrowtype='lefthalf',
                layer=cfg.bbox_pin_layer)
            nd.put_stub(['lc', 'tc', 'rc', 'bc'], arrowtype='full',
                layer=cfg.bbox_pin_layer)

        if name:
            cellname(cellname=_paramsname, length=length, align='cc').\
                put(bbox.pin['cc'])
        if params:
                parameters(parameters=_parameters).put(bbox.pin['cc'])

    # options to align the bbox in this cell w.r.t. to its pins:
    align_options = [ 'lb', 'lc', 'lt', 'rb', 'rc', 'rt', 'cc']
    align_translate = [
        (0, 0.5*width), (0, 0), (0, -0.5*width),
        (-length, 0.5*width), (-length, 0), (-length, -0.5*width),
        (-0.5*length, 0)
    ]

    dx = align_translate[align_options.index(align)][0]
    dy = align_translate[align_options.index(align)][1]

    box = bbox.put(cfg.cells[-1].pin[pinname].move(dx, dy).move(*move))

    if raise_pins:
        box.raise_pins(bbox_pinnames, show=False)

    return None


def load_gdsBB(gdsfilename, cellname, pinfilename=None, newcellname=None,
        layermap=None, flip=False, scale=1, stubs=True):
    """Load a GDS building block and its corresponding pin file.

    Args:
        gdsfilename (str): gds filename to import from
        cellname (str): cellname to import
        pinfilename (str): optional name of csv file containing pin info for stubs.
        newcellname (str): new cell name
        layermap (dict): mapping of layer {number_in: number_out}
        flip (bool): mirror the gds (default = False)
        scale (float): scaling factor (default = 1), not implemented
        stubs (bool): place stubs (default = True)

    Returns:
        Cell
    """
#    if stubs:
#        instantiate = True
#    else:
#        instantiate = False

    with nd.Cell(cellname+'_bb', instantiate=False) as C:
        #TODO: if this is true the instance and gds file may get the same name
        #      causing a topology error in Klayout
        nd.load_gds(
            filename=gdsfilename,
            cellname=cellname,
            newcellname=newcellname,
            layermap=layermap,
            scale=scale
            ).put(0, flip=flip)
        #TODO: if the GDS gets flipped as original the pins need to follow
        if pinfilename is not None:
            df = pd.read_csv(pinfilename, delimiter= ',', skiprows=1, header=None,
                names = ['io', 'x', 'y', 'a', 'xs', 'w'])
            df.fillna(0.0, inplace=True)
            for row in df.itertuples():
                i, io, x, y, a, xs, w = row
                try:
                    xs = xs.strip()
                    if xs.upper() == 'NONE':
                        xs = None
                except:
                    xs = None
                if not isinstance(w, float):
                    w = 0.0

                nd.Pin(name=io, xs=xs, width=w).put(x, y, a)
                if stubs and xs is not None :
                    put_stub(io)
    return C


# =============================================================================
# predefined example cell for documentation examples
# =============================================================================
with nd.Cell('example_cell') as example:
    s = nd.strt(length=20).put(0)
    b = nd.bend(angle=45).put()
    nd.Pin('a0', pin=s.pin['a0']).put()
    nd.Pin('b0', pin=b.pin['b0']).put()


dir_path = os.path.dirname(os.path.abspath(__file__))
nazca_logo = load_gdsBB(
    gdsfilename=os.path.join(dir_path, 'nazca_logo.gds'),
    cellname='nazca',
    newcellname='nazca_logo')