#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#==============================================================================
# (c) 2017 Katarzyna Lawniczuk, Ronald Broeke, Xaveer Leijtens
#==============================================================================
"""
Pixapp Package template
"""

from itertools import count
import nazca as nd
import nazca.pdk_template as pdk
import nazca.geometries as geom
import nazca.cfg as cfg


#==============================================================================
# Package template
#==============================================================================
annotationlayer = 'AnnotationIO'
fiberIOlayer = 'FiberIO'
textheight = 0.15
textmove = -0.08
padtext = 30

# RF pads can be on all sides, but not staggered.
RFname = ['rfT{:02d}', 'rfB{:02d}', 'rfL{:02d}', 'rfR{:02d}']

# DC pads can be on all sides, possibly staggered (in two rows).
# First row
DC0name = ['dc0B{:02d}', 'dc0T{:02d}', 'dc0L{:02d}', 'dc0R{:02d}']
# Second row
DC1name = ['dc1B{:02d}', 'dc1T{:02d}', 'dc1L{:02d}', 'dc1R{:02d}']

class Package1():
    """Class containing a Pixapp DieLayout template.

    Pad specification values:
        None:  no pads, no arrows
        False: no pads, show arrows
        True:  all pads, show arrows
        list:  pads in list, show arrows
    """
    instnum = count()

    def __init__(self, name='Pixapp-Package1',
            length=4000, height=4000, cleave=100,
            DCminside=0, RFminside=0,
            DC0edge=100, DC1edge=100+150, DCpitch=150,
            RFedge=90, RFpitch=150,
            fiberareaL=None, fiberareaR=1000,
            DC0B=True,  DC0T=True,  DC0L=None, DC0R=None,
            DC1B=None, DC1T=None, DC1L=None, DC1R=None,
            RFB=None, RFT=None, RFL=None, RFR=None,
            DCpadcell=None, RFpadcell=None,
            textlayer=None
            ):
        """Construct a package object.

        Args:
            length (float): die length (after cleaving)
            height (float): die height (after cleaving)
            cleave (float): thickness of cleave border
            DCminside (int): x
            RFminside (int): x
            DC0edge (float): x
            DC1edge (float): x
            DCpitch (float): pitch for DC-pad placement
            RFedge (float): x
            RFpitch (float): pitch for RF-pad placement
            fiberareaL (x): x
            fiberareaR (x): x
            DC0B (bool): place 0th-row Bottom dc pads
            DC0T (bool): place 0th-row Top dc pads
            DC0L (bool): place 0th-row Left dc pads
            DC0R (bool): place 0th-row Right dc pads
            DC1B (bool): place 1st-row Bottom dc pads
            DC1T (bool): place 1st-row Top dc pads
            DC1L (bool): place 1st-row Left dc pads
            DC1R (bool): place 1st-row Right dc pads
            RFB (x): place Bottom rf pads
            RFT (x): place Top rf pads
            RFL (x): place Left rf pads
            RFR (x): place Right rf pads
            DCpadcell (x): x
            RFpadcell (x): x
            textlayer (str | int): mask layer for text output

        Returns:
            None

        Example:
            Create a package object, create a cell with its method and put it::

            import nazca as nd
            from nazca.demopackager.packages import Package1

            PACKAGE = Package1(name='Package-A1', length=5000, height=4000)
            package.cell().put(0)

            nd.export_plt()
        """

        self.name   = name
        self.length = length
        self.height = height
        self.cleave = cleave

        # Distance to edge (includes pad length)
        self.DC0edge = DC0edge
        self.DC1edge = DC1edge
        self.DCpitch = DCpitch

        # Distance to edge (includes pad length)
        self.RFedge = RFedge
        self.RFpitch = RFpitch

        # Specify which DC pads to put.
        self.DC0 = [DC0B, DC0T, DC0L, DC0R]
        self.DC1 = [DC1B, DC1T, DC1L, DC1R]

        # Specify which RF pads to put.
        self.RF = [RFB, RFT, RFL, RFR]

        # minimum distance to sides of chip (pads are centered)
        self.DCminside = DCminside
        self.RFminside = RFminside

        self.fiberareaL = fiberareaL
        self.fiberareaR = fiberareaR
        self.textlayer = textlayer

        self.DCpadcell = DCpadcell
        self.RFpadcell = RFpadcell

        if self.textlayer is None:
            self.textlayer = 'AnnotationIO'
        else:
            self.textlayer = textlayer

        self.arrow = pdk.make_arrow(layer=cfg.package_pin_layer)


    def die_size(self, length=4000, height=4000):
        """Set size of the package area."""
        self.length = length
        self.height = height


    def cell(self):
        """Create a Cell with DC and RF postitions.

        Returns:
            Cell: cell for packaging"""
        # Optical positions come from foundry.
        l = self.length - self.cleave
        h = self.height - self.cleave

        with nd.Cell(name='{}_{}'.format(self.name, next(self.instnum))) as C:

            # Number of DC pads
            nBT = int((l-2*self.DCminside)/self.DCpitch)
            nLR = int((h-2*self.DCminside)/self.DCpitch)
            nDC = [nBT, nBT, nLR, nLR]

            # tuple in pins: start position of dc pins on all four sides.
            dc0 = (
                nd.Pin('p0B').\
                    put(l/2-(nBT/2-0.5)*self.DCpitch, self.DC0edge, 90), # Bottom
                nd.Pin('p0T').\
                    put(l/2+(nBT/2-0.5)*self.DCpitch,  h-self.DC0edge, 270), # Top
                nd.Pin('p0L').\
                    put(self.DC0edge, h/2+(nLR/2-0.5)*self.DCpitch, 0), # Left
                nd.Pin('p0R').\
                    put(l-self.DC0edge, h/2-(nLR/2-0.5)*self.DCpitch, 180) # Right
            )

            # Number of RF pads
            nBT = int((l-2*self.RFminside)/self.RFpitch)
            nLR = int((h-2*self.RFminside)/self.RFpitch)
            nRF = [nBT, nBT, nLR, nLR]

            # tuple in pins: start position of dc pins on all four sides.
            rf0 = (
                nd.Pin('rfB').\
                    put(l/2-(nBT/2-0.5)*self.RFpitch, self.RFedge, 90), # Bottom
                nd.Pin('rfT').\
                    put(l/2+(nBT/2-0.5)*self.RFpitch, h-self.RFedge, 270), # Top
                nd.Pin('rfL').\
                    put(self.RFedge, h/2+(nLR/2-0.5)*self.RFpitch, 0), # Left
                nd.Pin('rfR').\
                    put(l-self.RFedge, h/2-(nLR/2-0.5)*self.RFpitch, 180) # Right
            )

            # Place DC0 pins
            for pin, n, name, pad in zip(dc0, nDC, DC0name, self.DC0):
                if pad is None:
                    continue # Don't generate the pins
                for i in range(0, n):
                    loc = nd.Pin(name.format(i)).\
                        put(pin.move(0,-i*self.DCpitch,0))
                    self._put_arrow(loc, name, i)
                    if pad is False or self.DCpadcell is None:
                        continue # Don't draw the pad
                    if pad is True or i in pad:
                        self._put_pad(loc, self.DCpadcell, name, i)
                        # Set properties of the template pin.
                        loc.xs = self.DCpadcell.pin['c0'].xs
                        loc.width = self.DCpadcell.pin['c0'].width

            # Place DC1 pins
            for pin, n, name, pad in zip(dc0, nDC, DC1name, self.DC1):
                if pad is None:
                    continue # Don't generate the pins
                for i in range(0, n-1):
                    loc = nd.Pin(name.format(i)).\
                        put(pin.move(self.DC1edge-self.DC0edge,
                            -(i+0.5)*self.DCpitch, 0))
                    self._put_arrow(loc, name, i)
                    if pad is False or self.DCpadcell is None:
                        continue # Don't draw the pad
                    if pad is True or i in pad:
                        self._put_pad(loc, self.DCpadcell, name, i)
                        # Set properties of the template pin.
                        loc.xs = self.DCpadcell.pin['c0'].xs
                        loc.width = self.DCpadcell.pin['c0'].width


            # Place RF pins
            for p, n, name, pad in zip(rf0, nRF, RFname, self.RF):
                if pad is None:
                    continue # Don't generate the pins
                for i in range(0, n):
                    loc = nd.Pin(name.format(i)).\
                        put(p.move(0,-i*self.RFpitch,0))
                    self._put_arrow(loc, name, i)
                    if pad is False or self.RFpadcell is None:
                        continue # Don't draw the pad
                    if pad is True or i in pad:
                        self._put_pad(loc, self.RFpadcell, name, i)
                        # Set properties of the template pin.
                        loc.xs = self.RFpadcell.pinin.xs
                        loc.width = self.RFpadcell.pinin.width

            #Fiber positions
            faalen = self.cleave+50
            if self.fiberareaR:
                dist_edgey = self.height-self.cleave-self.fiberareaR
                farea = geom.box(width=self.fiberareaR, length=faalen)
                poly = nd.Polygon(layer=fiberIOlayer, points=farea)
                poly.put(self.length-0.5*self.cleave,
                    0.5*dist_edgey+0.5*self.fiberareaR, 0)
            if self.fiberareaL:
                dist_edgey = self.height-self.cleave-self.fiberareaL
                farea = geom.box(width=self.fiberareaL, length=faalen)
                poly = nd.Polygon(layer=fiberIOlayer, points=farea)
                poly.put(-faalen-0.5*self.cleave,
                    0.5*dist_edgey+0.5*self.fiberareaL, 0)

        return C


    @property
    def maxDCcount(self):
        """Get max number of DC ports that fit in a package"""
        return round((self.length-2*self.DCx)/self.DCpitch)


    @property
    def maxRFcount(self):
        """Get max number of RF ports that fit in a package"""
        return round((self.length-2*self.RFy)/self.RFpitch)


    def _put_pad(self, loc, pad, name, i):
        pp = pad.put(loc.rotate(180))
        pname = name[:-7] + '\n' + name[-7:].format(i)
        t = nd.text(pname, height=padtext, align='cc', layer=annotationlayer)
        t.put(pp.pinin.move(-45, 0, 270))

    def _put_arrow(self, loc, name, i):
        self.arrow.put(loc)
        t = nd.text(name.format(i), layer=annotationlayer,
            height=textheight, align='rc')
        t.put(loc.move(textmove))
