#!/usr/bin/env python3
#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#
# Character output
#
# (c) 2015 Xaveer Leijtens
#
#
from .font import font
from .netlist import put_polygon, put_polyline, Cell, Pin
from .util import md5
from . import cp
from .mask_layers import get_layer
import warnings

lwf = 0.07   # linewidth as fraction of height
lh   = 5     # logical height of character (capital + descender)
lss  = 1.06  # line spacing scale
sp   = 1.5   # intercharacter space
wh   = 4.5/8 # width to height ratio


def get_lw(height, stretch):
    """Returns:
        float: linewidth of character stroke
    """
    return min(lwf, lwf * stretch) * height


def textheight(text, height=50):
    """Returns:
        float: total textheight of line(s) in text
    """
    return (text.count('\n') * lss + 1) * height


def linelength(text, height=50, stretch=1):
    """Returns:
        float: total linelength of longest line in (multiline) text
    """
    maxl = 0
    for line in text.split('\n'):
        # return width of line
        l = ((sum(font[char][0] for char in line) +
            (len(line) - 1) * sp) * wh / lh) * height * stretch + \
            get_lw(height, stretch)
        maxl = max(maxl, l)
    return l


def text(text="Text", height=50, layer=1, stretch=1, align='lb'):
    """Output a cell with text of specified dimensions and alignment.

    Args:
        text (str): text to display
        height (float): height off the text in um
        layer (int | string): layer number or name to place text in
        stretch (float): (default = 1.0)
        align (str): relative placement: regex: [lcr][bct] (default = 'lb')
            examples: 'lc', 'cc', 'tr'

    Returns:
        Cell: cell with text as provided in <text>

    Example:
        >>> message = text(text='Hello world', align='cc')
        >>> message.put()
    """

    if isinstance(layer, str):
        layer = get_layer(layer)

    halign, valign = align.lower()
    if halign not in 'lcr':
        warnings.warn("text halign parameter not one of lcr, default to 'l'.")
        halign = 'l'
    if valign not in 'bct':
        warnings.warn("text valign parameter not one of bct, default to 'b'.")
        valign = 'b'

    # Character parameters
    lw = get_lw(height, stretch)
    cheight = height - lw # height of characters w/o line thickness
    xscale = height * wh / lh * stretch
    yscale = cheight / lh
    ls = lss * height # line spacing

    id5 = md5(text + align.lower() + "{}{}{}".format(height, layer, stretch))
    with Cell('text_' + id5, instantiate=False) as txt:
        bby = textheight(text, height)
        if valign == 'c':
            ypos = bby / 2 - cheight
        elif valign == 't':
            ypos = -cheight
        else:
            ypos = bby - cheight
        for line in text.split('\n'):
            bbx = linelength(line, height, stretch)
            if halign == 'c':
                xpos = -bbx / 2
            elif halign == 'r':
                xpos = -bbx
            else:
                xpos = 0
            for char in line:
                w, strokes = font[char]
                charwidth = w * xscale
                for stroke in strokes:
                    XY = [(x * xscale, y * yscale) for x, y in stroke]
                    if len(XY) > 1:
                        put_polyline(connect=(xpos + lw/2, ypos - lw/2, 0),
                                points=XY,
                                width=lw,
                                layer=layer,
                                pathtype=2)
                xpos = xpos + charwidth + sp * xscale
            ypos -= ls
        Pin(name='a0', xs=None).put(0, 0, 180)
        Pin(name='b0', xs=None).put(xpos, ypos+ls-lw, 0)
    return txt

Text = text
