#!/usr/bin/env python3
#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#
# Polygon font output
#
# (c) 2017 Xaveer Leijtens
#
#
import pickle
import warnings
import os
from .netlist import put_polygon, Cell, Pin
from .util import md5
from .mask_layers import get_layer

# Nazca font files have the following structure
# version: font file version number
# height: total height of the font
# space: interword space (=0 for fixed fonts)
# font: dictionary of characters:
#   key: character
#   w: width of that character
#   p: list of polygons, of which each is a list of (x,y) coordinate points

lss  = 1.06  # line spacing scale

# Path to this file
modulepath = os.path.dirname(os.path.abspath(__file__))

class textfont():
    def __init__(self, fontfile='cousine.nft'):
        ff = fontfile
        if not os.path.exists(ff):
            ff = os.path.join(modulepath, ff)
            if not os.path.exists(ff):
                print("Can't open font '{}'. Using 'cousine.nft'.".format(
                    fontfile))
                ff = os.path.join(nd.modulepath, 'cousine.nft')
        with open(ff, 'rb') as f:
            self.version, self.lh, self.spc, self.tfont = pickle.load(f)

    def textheight(self, text, height=50):
        """Return total textheight of line(s) in text.
        """
        return (text.count('\n') * lss + 1) * height

    def linelength(self, text, height=50, spc=0):
        """Return total linelength of longest line in (multiline) text.
        """
        maxl = 0
        for line in text.split('\n'):
            # return width of line
            l = sum(self.tfont[char][0] for char in line) * height / self.lh
            l += (len(line) - 1) * spc
            maxl = max(maxl, l)

        return l

    def text(self, text="Text", height=50, layer=1, align='lb'):
        """Output text of specified dimensions and alignment.
        """

        if isinstance(layer, str):
            layer = get_layer(layer)

        halign, valign = align.lower()
        if halign not in 'lcr':
            warnings.warn("text halign parameter not one of lcr, default to 'l'.")
            halign = 'l'
        if valign not in 'bct':
            warnings.warn("text valign parameter not one of bct, default to 'b'.")
            valign = 'b'

        # Character parameters
        scale = height / self.lh
        ls = lss * height # line spacing

        id5 = md5(text + align.lower() + "{}{}".format(height, layer))
        with Cell('textf_' + id5, instantiate=False) as txt:
            bby = self.textheight(text, height)
            if valign == 'c':
                ypos = bby / 2 - height
            elif valign == 't':
                ypos = -height
            else:
                ypos = bby - height
            for line in text.split('\n'):
                bbx = self.linelength(line, height, self.spc)
                if halign == 'c':
                    xpos = -bbx / 2
                elif halign == 'r':
                    xpos = -bbx
                else:
                    xpos = 0
                for char in line:
                    w, polys = self.tfont[char]
                    for poly in polys:
                        XY = [(x * scale, y * scale) for x, y in poly]
                        if len(XY) > 3:
                            put_polygon(connect=(xpos, ypos, 0),
                                    points=XY,
                                    layer=layer)
                    xpos = xpos + w * scale + self.spc
                ypos -= ls
            Pin(name='b0', xs=None).put(xpos, ypos+ls, 0)
        return txt
