#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------
#
# Utility routines
#
# (c) 2016 Xaveer Leijtens
#
from collections import OrderedDict
from . import gds_base as gbase
from .netlist import Cell
from PIL import Image
import os
import numpy as np
import hashlib

__all__ = ['parameters_to_string', 'string_to_parameters',
        'get_cell_annotation', 'get_cell_polyline', 'get_cell_polygon',
        'md5', 'isnotebook', 'image']

def parameters_to_string(param):
    # param is a dictionary with parameter name as key.
    plist = ['Parameters:']
    for key, value in param.items():
        plist.append("{} = {}".format(key, value))
    return '\n'.join(plist)

def string_to_parameters(string):
    # Return ordered dictionary containing the parameters.
    lines = string.split('\n')
    p = OrderedDict()
    if (lines[0] == 'Parameters:'):
        for line in lines[1:]:
            param = line.split('=', 1)
            if len(param) == 2:
                p[param[0].strip()] = param[1].strip()
    return p

def get_cell_annotation(cell, convert=False):
    # If not convert then return XY as integers, as in GDSII file (nm).
    # If convert then return XY as floats (um).
    """Yield annotation
    layer, position, text
    """
    if convert:
        conv = gbase.gds_db_user
    else:
        conv = 1
    for e in cell.elements:
        if e.etype == gbase.gds_record.TEXT:
            lay, pos, text = e.annotation
            pos[0] *= conv
            pos[1] *= conv
            yield lay, pos, text

def get_cell_polyline(cell, convert=False):
    # If not convert then return XY as integers, as in GDSII file (nm).
    # If convert then return XY as floats (um).
    """Yield polylines
    layer, XY
    """
    if convert:
        conv = gbase.gds_db_user
    else:
        conv = 1
    for e in cell.elements:
        if e.etype == gbase.gds_record.PATH:
            lay, points = e.polyline
            XY = []
            for i in range(0, len(points), 2):
                XY.append((points[i] * conv, points[i+1] * conv))
            yield lay, XY

def get_cell_polygon(cell, convert=False):
    # If not convert then return XY as integers, as in GDSII file (nm).
    # If convert then return XY as floats (um).
    """Yield polygons
    layer, XY
    """
    if convert:
        conv = gbase.gds_db_user
    else:
        conv = 1
    for e in cell.elements:
        if e.etype == gbase.gds_record.BOUNDARY:
            lay, points = e.polygon
            XY = []
            for i in range(0, len(points), 2):
                XY.append((points[i] * conv, points[i+1] * conv))
            yield lay, XY

def md5(s, N=5):
    # Return first N characters of md5 hash of string s.
    return hashlib.md5(s.encode()).hexdigest()[:N]

def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':  # Jupyter notebook or qtconsole?
            return True
        elif shell == 'TerminalInteractiveShell':  # Terminal running IPython?
            return False
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

def PIL2array(img):
    return np.array(img.getdata(),
            np.bool).reshape(img.size[1], img.size[0])

def image(name, layer=1, size=256, pixelsize=1, threshold=0.5,
        cellname=None, invert=False, align='cc'):
    """Read an image file and return a nazca cell with the image.

    Image format can be png, jpg, gif, bpm, eps and others, as supported by Pillow.

    Args:
        name (str): name of the image file
        layer (int): layer number that the image will be written to (default 1)
        size (int): maximum bounding box size in pixels (default 256)
        pixelsize (float): pixel size in micron (default 1)
        threshold (float): black/white threshold (default 0.5)
        cellname (str): nazca cell name (default image filename)
        invert (bool): flag to invert black & white (default False)
        align (str): two character string for image alignment (default 'cc')
            allowed:
            lt, ct, rt,
            lc, cc, rc,
            lb, cb, rb

    Returns:
        Cell: cell with image

    Examples:
        Load a logo in a cell and put it.

        >>> logo = image('mylogo.png', align='lb') # left/bottom alignment
        >>> logo.put()
    """
    if cellname is None:
        cellname = os.path.basename(name)
    p = pixelsize
    threshold = int(threshold * 256)
    a = {'lb', 'cb', 'rb', 'lc', 'cc', 'rc', 'lt', 'ct', 'rt'}
    if align not in a:
        print("Invalid alignment specification '{}' for image '{}'.".format(align, name))
        print("Allowed values are {}.".format(a))
        print("Using default value 'cc'")
        align = 'cc'
    halign = {'l': 0, 'c': -0.5, 'r': -1}
    valign = {'b': 0, 'c': -0.5, 't': -1}

    im = Image.open(name)
    gray = im.convert('L')
    # resize keep aspect, only if smaller
    gray.thumbnail((size, size), Image.ANTIALIAS)
    bw = gray.point(lambda x: 0 if x<threshold else 255, '1')
    pix = PIL2array(bw)
    width, height = bw.size
    print('Generating {}x{} image'.format(width, height))
    h0 = halign[align[0]] * width * p
    v0 = valign[align[1]] * height * p

    with Cell(cellname) as C:
        for line in range(height):
            x1 = x0 = lb = lw = 0
            y0 = (height - line) * p
            for pixel in pix[line]:
                if pixel == invert:
                    lb += 1
                    if lw > 0:
                        x0 += lw * p
                        lw = 0
                else:
                    lw += 1
                    if lb > 0:
                        x1 = x0 + lb * p
                        xy = [(x0, y0), (x1, y0), (x1, y0-p), (x0, y0-p)]
                        polygon(layer=layer, points=xy).put(h0, v0, 0)
                        x0 = x1
                        lb = 0
            if lb > 0:
                x1 = x0 + lb * p
                xy = [(x0, y0), (x1, y0), (x1, y0-p), (x0, y0-p)]
                polygon(layer=layer, points=xy).put(h0, v0, 0)
    return C


def klayout2nazca(string):
    """Convert copy-pasted path from KLayout into list of points.

    Args:
        string (str): string with points to convert

    Return:
        list of points (x,y)

    Example:

        >>> klayout2nazca("1.0, 3.5, 2.0, 5.5")
        [(1.0, 3.5), (2.0, 5.5)]
    """
    return [[float(x) for x in line.split('\t')]
        for line in string.strip().split('\n')]